/*****************************************************************************
 * NCSA Polyview 3.1                                                         *
 *                                                                           *
 * Version 3.1 changes and additions by Gilles Bourhis.                      *
 * Version 3 changes and additions by Marc Andreessen.                       *
 * Version 2 by Brian Calvert.                                               *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvdata.h,v 1.3 93/08/13 12:59:06 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvdata.h,v $
 * Revision 1.3  93/08/13  12:59:06  gbourhis
 * Add rgb_box field in polyview_s structure.
 * 
 * Revision 1.2  93/07/13  16:33:08  gbourhis
 * use int32 for file handle instead of FILE *.
 * 
 * Revision 1.1  92/09/18  10:55:26  marca
 * Initial revision
 * 
#endif

#ifndef __POLYVIEW_PVDATA_H__
#define __POLYVIEW_PVDATA_H__

/* This file should be included after files containing definitions
   of datatypes used herein. */

/* ------------------------------- pickmode ------------------------------- */
typedef enum 
{
  PICK_NONE, PICK_VERTEX, PICK_EDGE, PICK_FACE, 
  PICK_OBJECT, PICK_BELOW, PICK_ALL
} pickmode_t;


/* ----------------------------- parse & arg ------------------------------ */
/* parse_t defines an action to the parser.  The format describes the syntax
   of the action.  action_fn is a function that accepts the system state,
   pointer to an action, and active window pointer. */
typedef struct parse_s 
{
  char format[MAXFORMATLEN];
  int immediate;              /* like "abort", "exit", "move" */
  int global;                 /* TRUE if independent of active */
                              /* window, FALSE otherwise. */
  int (*action_fn)();
} parse_t;

typedef struct arg_s 
{
  int type;
  char *description;
  char *value;
} arg_t;


/* -------------------------------- stats --------------------------------- */
/* stats_t describes the statistics for an associated vdata. */
typedef struct stats_s 
{
  /* type indicates what type of data this data set is.  The only */
  /* types presently supported are: INTEGER and FLOAT. */
  int type;
  
  /* rec_count is the number of records in the vdata. */
  int rec_count;
  
  /* max_rec_count is the maximum number of records in the vdata. */
  /* Primarily useful when working with dynamic data (such as that */
  /* from a DTM port). */
  int max_rec_count;
  
  /* rec size is the number of elements in each record.  Since all */
  /* vdatas are homogenous at this point, rec_size * rec_count *   */
  /* sizeof(datatype) indicates the size of the space needed to store */
  /* this vdata. */
  int rec_size;
  
  /* min is the minimum value in the associated vdata. */
  /* max is the maximum value in the associated vdata. */
  float min, max;
    
  /* mean is the value sum(vdata)/(rec_count*rec_size). */
  /* mean is the value sum(vdata*vdata)/(rec_count*rec_size). */
  /* variance is an estimate of the vdata's variance made using the */
  /* equation mean_sq-(mean*mean). */
  float mean, mean_sq, variance;
    
} stats_t;


/* Forward typedef for files_t. */
typedef struct files_s files_t;

/* -------------------------------- Vdata --------------------------------- */
/* Data sharing is a key feature of memory conservation in this program. */
/* The main place this occurs is in vdatas.  If two vdatas have the same */
/* ref value in the file, they contain the same data.  In this memory-bound */
/* copy of the HDF vset structure, the data pointed to by the data pointer */
/* is shared, when possible, by identical vdatas in the same file.   The */
/* vdata record itself, however, is not shared. */
typedef struct Vdata_s 
{
  /* id is the reference identifier associated with this vdata in the */
  /* HDF file.   If this record is not associated with an HDF file, */
  /* this value is -1. */
  int id;
  
  /* name is the name of the vdata associated with this record. */
  char name[MAXNAMELEN];
  
  /* fields is the list of fields associated with this record. */
  char fields[MAXNAMELEN];
  
  /* file is a pointer back to the file that this vdata is a member */
  /* of.  This allows us to easily open the file if necesary to read */
  /* this data back in. */
  files_t *file;
  
  /* locked is TRUE if the block is not to be freed up to make room */
  /* for a new vdata to be read into memory.  Initially FALSE, this */
  /* field is provided so that the used can specify important or */
  /* sizeable data blocks that should not be paged out.  If there */
  /* is no room to free up anywhere else, the user will be prompted */
  /* by the swap_out routine as to whether to override the lock. */
  int locked;
  
  /* in memory is TRUE if the data associated with this vdata is */
  /* resident in memory at this point.  Reset to FALSE if the data */
  /* block needs to be freed up to make room for other mallocs. */
  int in_memory;
  
  /* last used is set to the last state_t->event_count that this */
  /* vdata was used.  The value is initially zero and is set to the */
  /* current event_count whenever it is read.  This value is used when */
  /* determining what blocks need to be deleted to make room for other */
  /* vdatas. */
  long last_used;
  
  /* stats read is TRUE if the statistics for this data set have */
  /* already been read.  Using this flag, recalculation of statistics */
  /* may be avoided. */
  int stats_read;
  
  /* stats contains statistics describing the data associated with */
  /* this record. */
  stats_t stats[MAXDATADIMS];
  
  /* data points to the data associated with this vdata record. */
  char *data;
  
  /* sibling points to the next vdata at this level. */
  struct Vdata_s *sibling;
  
  /* next_pool points to the next vdata in the pool for this file. */
  struct Vdata_s *next;
} Vdata_t;


/* -------------------------------- Vgroup -------------------------------- */
typedef struct Vgroup_s {
  /* name is the name of the Vgroup associated with this record. */
  char name[MAXNAMELEN];
  
  /* id is the integer id associated with this vgroup */
  int id;
  
  /* file is a pointer back to the file that this vgroup is a member */
  /* of. */
  files_t *file;
  
  /* children read is TRUE if the children of this group have already */
  /* been transferred into memory. */
  int children_read;
  
  /* sibling points to the next group record in the current level */
  /* of the tree. */
  struct Vgroup_s* sibling;
  
  /* groups points to the child groups in the current context.  If */
  /* there are no sub-Vgroups, or none of them have been read in */
  /* yet, the pointer is NULL. */
  struct Vgroup_s* groups;
  
  /* vdatas points to the vdatas in the current context.  If there */
  /* are no vdatas, or none of them have been read in yet, the */
  /* pointer is NULL. */
  Vdata_t *vdatas;
} Vgroup_t;


/* -------------------------------- files --------------------------------- */
struct files_s 
{
  /* type is the type of file referred to.  At this point, RIS8, HDF, */
  /* TEXT, and VSET are supported. */
  int type;
  
  /* path is the pathname of the file associated with this record. */
  char path[MAXPATHLEN];
  
  /* file fh is the file handle to file, if it is open.  Otherwise */
  /* it is NULL. */
  int32 fh;
  
  /* port is the id of the port associated with this "file" if it is */
  /* a socket. */
  int portid;
  
  /* next points to the next file record in the tree. */
  files_t *next;
  
  /* groups points to the top-level group in a Vset file.  This is */
  /* always a "root Vgroup" which has an empty name.  This simplifies */
  /* the traversal routines since they can always start with one root */
  /* vgroup. */
  Vgroup_t *groups;
  
  /* vdatas points to the list of all vdatas in the file.  If none */
  /* exist in memory yet, this pointer is NULL. */
  Vdata_t *vdatas;
};


/* ------------------------------- palette -------------------------------- */
/* palette_t describes the information associated with a given palette.  */
/* This data is typically associated with some kind of data-viewing window */
/* and is loaded into the colormap when that window is displayed. */
typedef struct palette_s {
  /* name is the unique name of this palette.  Windows associated */
  /* with this palette name will be directly linked into this record. */
  char name[MAXNAMELEN];
  
  /* refs is a count of the number of window making references to this */
  /* palette.  If refs is zero, nothing is referencing this palette. */
  /* The value is incremented every time this palette is added to a */
  /* structure. */
  int refs;
  
  /* base is the base entry of this palette in the colormap range of */
  /* 0..4095. */
  int base;
  
  /* slope and origin describe the mapping function between the */
  /* master palette and the display palette. */
  float slope;
  float origin;
  
  /* slopestep and originstep control the amount of change in the slope */
  /* and origin, respectively, when the palette is annimated. */
  float slopestep;
  float originstep;
  
  /* master palette conatins the standard, non-shifted mapping between */
  /* normalized data values and RGB color components. */
  short master_palette[256][3];
  
  /* display palette is the modified master palette that is */
  /* manipulated by the mapping function. */
  short display_palette[256][3];
  
  /* probe value is the color map index to the currently probed color. */
  /* If this value is negative, no color is currently selected. */
  int probe_value;
  
  /* min and max are the minimum and maximum scientific values */
  /* corresponding to the palette range (base+1, base+PALETTESIZE-2). */
  /* They are used for calculating which scientific values correspond */
  /* to a given color. */
  float min;
  float max;
} palette_t;


/* Forward pointers; may not be necessary. */
typedef struct window_s window_t;
typedef struct frame_s frame_t;
typedef struct objinfo_s objinfo_t;

/* -------------------------------- probe --------------------------------- */
typedef struct probe_s {
  /* type is the kind of object that is selected for probing.  It */
  /* could be a NODE, ELEMENT, or GROUP. */
  int type;
  
  /* id is a unique identifier for the selected object.  It is used to */
  /* index into the proper data elements so that correct values may be */
  /* displayed. */
  long id;
  
  /* probed window indicates the polyview window that this window is */
  /* probing into.  */
  window_t *probed_window;
  
  /* next points to the next probe window associated with this */
  /* polyview window.  This chain is used to update all affected probe */
  /* windows when a polyview window is updated. */
  window_t *next;
} probe_t;


/* -------------------------------- window -------------------------------- */
/* window_t contains information describing a window. */
struct window_s {
  /* NEW FIELDS: */
  /* uniqid uniquely identifies this window and its Motif hierarchy
     in relation to all other windows.  When we get a callback from a
     Motif widget we extract the uniqid from the client_data and use
     that to find the window in which the action was triggered. */
  /* NO NO NO!  Use 'id' below instead; that way you don't have
     to change all the names, idiot. */
  /* int uniqid; */
  
  /* base is the base Widget for this window and its hierarchy.
     It is generally the dialog itself, or the toplevel widget for
     the main (status) window. */
  Widget base;
  
  /* string in which to save debugging string */
  char *debug_str;
  
  /* TEMPORARY (hopefully) fields for initial parameters when
     doing real-time rotation/translation with mouse. */
  /* These are set initially when button2 goes down, and then
     modified as button2motion happens.  We know we won't ever
     end up using stale values since button2 always goes down before
     button2motion happens again. */
  float rho, theta, phi;
  float rho_v, theta_v, phi_v;
  float CI[DIMS];
  float half, radius;
  int shifted, shift_changed;
  
  /* WINDOW INFO: */
  /* id is the unique identifier of the window associated with this */
  /* record.  The value is used to focus GL's attention on the proper */
  /* window during drawing operations. */
  int id;
  
  /* name is the name of the window.  It is displayed on the top of */
  /* the window frame.  It should not change once established, and */
  /* must be unique among windows of its type. */
  char name[MAXNAMELEN];
  
  /* type indicates what kind of window this is.  It may, for example, */
  /* be a POLYVIEW, PALETTE, STATUS, or PROBE window. */
  int type;
  
  /* root_menu is the identifier of the root menu for this window. */
  /* This value may be changed to different values.  If the id is */
  /* non-negative, it is used to post a menu whenever the right mouse */
  /* button is pressed inside this window. */
  int root_menu;
  
  /* DATA: */
  /* image points to a record containing the information necessary */
  /* to render the window's image.  The structure of the record is */
  /* determined by type.  This pointer may be NULL for some window */
  /* types. */
  /* THIS IS VOID NOW. */
  void *image;
  
  /* info points to the root viewing information record.  This record */
  /* is used for all frames which are children of this window. */
  objinfo_t *info;
  
  /* frames is the number of frames in the sequence presently */
  /* associated with the window.  frame id is the sequence number of */
  /* the current frame, which is in the range [1..frames]. */
  int frames;
  int frame_id;
  
  /* frame_list points to the first in a sequence of frames associated */
  /* with this window.  frame points to the one currently being */
  /* displayed. */
  frame_t *frame_list;
  frame_t *frame;
  
  /* palette points to the palette data record associated with this */
  /* window.  Whenever the window is entered, this palette should be */
  /* swapped in. */
  palette_t *palette;
  char inpalette[MAXPATHLEN];
  char outpalette[MAXPATHLEN];
  
  /* RASTER IMAGES */
  /* rastername is the name of the file to which raster output is */
  /* to be sent. If no ".hdf" extension is provided, one is appended */
  /* by routines that use this.  This also serves as a base name for */
  /* series of raster files. */
  char rastername[MAXPATHLEN];
  
  /* rastercont indicates whether or not a continuous series of raster */
  /* images are to be output to the disk or not.  When TRUE, */
  /* continuous output is enabled. */
  int rastercont;
  
  /* Same as above, but for DTM. */
  int rasterblastcont;
  
  /* rasternum is the sequence number of the next image to be written */
  /* to the disk.  This is used whenever an image is written, and */
  /* should be updated accordingly. */
  int rasternum;
  
  /* source points to the data source associated with this window. */
  files_t *source;

#ifdef PRIVATE_SUPPORT
  /* This is a bad place for it, but...... is this window private?
     (private == no DTM communication allowed) */
  int private;
#endif
  
  /* vdata statistics for the entire sequence associated with the */
  /* polyview window. */
  stats_t stats[MAXDATADIMS][MAXVDATAS];
  
  /* vname is an array of the vdata names used for each of the data */
  /* slots in the image. */
  char vname[MAXVDATAS][MAXNAMELEN];
  
  /* vgroup is the name of the vgroup associated with this window
     (I think). */
  char vgroup[MAXPATHLEN];
  
  /* group ptr points to the Vgroup_t record corresponding to the */
  /* active group.  The record contains the name of the group.  If the */
  /* pointer is NULL, there is no active group. */
  Vgroup_t *vgroup_ptr;
  
  /* ANIMATION PARAMS: */
  /* animated is TRUE if the window's frame has already been advanced */
  /* somewhere in the process_events loop.  It is FALSE if it has not */
  /* been advanced yet.  This prevents us from skipping frames in the */
  /* sequence. */
  int animated;
  
  /* countdown is the number of frames that still need to be skipped */
  /* before the frame pointer is advanced to the next frame.  If zero, */
  /* the next frame is immediately grabbed and the value reinitialized */
  /* to the value of speed. */
  int countdown;
  
  /* speed indicates the number of TIMER0 interrupts to be ignored */
  /* between frames.  0 indicates maximum speed, any number other than */
  /* zero indicates the number of frames that will use the same data */
  /* before the next frame in the sequence is used.  If the value is */
  /* negative, it indicates that the window is not animating. */
  int speed;
  
  /* skip indicates the number of frames to skip between data updates. */
  /* To display every frame, skip should be 0.  To display every other */
  /* frame, slip should be 1, and so on. */
  int skip;
  
  /* reverse, if positive, indicates the number of frames that are to */
  /* be animated backwards before stopping.  If zero, no backwards */
  /* animation is taking place.  If negative, the animation is to */
  /* continue indefinately. */
  int reverse;
  
  /* forward, if positive, indicates the number of frames that are to */
  /* be animated forward before stopping.  If zero, no forward */
  /* animation is taking place.  If negative, the animation is to */
  /* continue indefinately. */
  int forward;
  
  /* 1 if forward/backward animation should bounce, 0 otherwise. */
  int bounce;
  
  /* step is the movement size to take for each frame of a move */
  /* command. */
  float step;
  
  /* movepath describes the path to be taken in any move commands. */
  /* Both spherical and cartesian interpolation may be used. */
  int movepath;
  
  
  /* WINDOW PARAMS: */
  /* please_redraw is TRUE when something has happened within the */ 
  /* process_events loop to cause us to redraw the window.  It is */
  /* reset to FALSE initially at the beginning of the loop. */
  int please_redraw;
  
  /* fixed is TRUE if the size of the window is locked.  If the window */
  /* size is locked, there is no need to update the location or size */
  /* when redrawing the window.  This flag is set when the window is */
  /* created. */
  int locked;
  
  /* freeze is TRUE when the window is frozen from updates.  This only */
  /* time a frozen window will be redrawn is when necessitated by a */
  /* portion of the window being exposed. */
  int freeze;
  
  /* x and y are the screen coordinates of the lower left corner of */
  /* the lower left corner of the window.  These values are updated */
  /* after any window redraw event. */
  long x, y;
  
  /* width and height are the dimensions of the window in terms of */
  /* screen coordinates.  These values are updated after any window */
  /* redraw event. */
  long width, height;
  
  /* cx and cy are the screen coordinates of the center of the window. */
  /* This value is updated after any window redraw event. */
  int cx, cy;
  
  /* bits is either 8 or 24, which indicates which type of graphics */
  /* display was initialized for this window. */
  int bits;
  
  /* shade indicates the type of shading to be applied to polygons. */
  /* The value of this field is either FLAT or GOURAUD. */
  long shade;
  
  /* light indicates whether or not the light is turned on */
  /* The value of this field is either 0 or 1. */
  int light;
  
  /* light_x, _y, and _z are the x, y, and z values of the 
     light vector */
  float light_x, light_y, light_z;
  
  /* light_twoside indicates whether or not twoside lighting
     is enabled */
  int light_twoside;
  
  
  /* draw a palette legend in the window? */
  int legend_flag;
  
  /* draw polyview credits in the window? */
  int credits_flag;
  
  /* outline_color is the color used for the outline when drawing */
  /* outlined faces. */
  int outline_color;
  
  /* line_width is width of lines in edges and outlined faces
     drawing modes. */
  int line_width;
  
  /* radius of spheres, or 0.0 if points should be drawn */
  float sphere_size;
  
  /* depth of sphere tesselation, according to sphere library */
  int sphere_depth;
  
  /* backdrop - background rectangle */
  int backdrop;
  
  /* VIEW PARAMS: */
  /* move describes whether the camera, reference point, or clipping */
  /* plane is currently under control of the mouse. */
  int move;
  
  /* projection is the projection type currently used to display this */
  /* window (PERSPECTIVE or ORTHOGONAL). */
  int projection;
  
  /* cfrom and sfrom are the cartesian and spherical representations, */
  /* respectively, of the camera location.  Spherical values are */
  /* measured relative to the reference point. */
  Coord cfrom[DIMS];
  Coord sfrom[DIMS];
  
  /* at is the cartesian location of the reference point for this */
  /* window. */
  Coord at[DIMS];
  
  /* fov_angle is the vield of view angle in tenths of degrees. */
  Angle view_angle;
  
  /* twist is the viewing twist angle. */
  Angle twist;
  
  /* twist_tmp is Marc's hack */
  Angle twist_tmp;
  
  /* up is a vector which determines the "up" direction of the camera. */
  /* The top of the view will be oriented in the direction of the */
  /* up vector. */
  Coord up[DIMS];
  
  /* max_width is the maximum width of the current objects as measured */
  /* from the minimum values corner to the maximum values corner. */
  float max_width;
  
  /* The picking mode currently used.  Picked object lists are */
  /* contained with the objects and are NULL pointers if nothing is */
  /* picked and a buffer of the appropriate size for the picking mode. */
  pickmode_t pickmode;
  
  
  
  /* FUNCTIONS: */
  /* init_fn is the function that initializes or resets the window */
  /* structure.  It is called with the following parameters: */
  /*    (*init_fn)(state_t *state, window_t *window, int reset) */
  /* where reset is FALSE if the window is new, TRUE if the window */
  /* is to be reinitialized. */
  int (*init_fn)();
  
  /* redraw_fn is the function to be invoked in order to redraw this */
  /* window.   It is called with the following parameters: */
  /*    (*redraw_fn)(state_t *state, window_t *window) */
  int (*redraw_fn)();
  
  /* animate_fn is the function to be invoked in order to update this */
  /* window's parameters for the next step of an animation sequence. */
  /* It is called with the following parameters: */
  /*    (*animate_fn)(state_t *state, window_t *window) */
  int (*animate_fn)();
  
  /* notify_fn is the function to be invoked when a mesage is given to */
  /* this window.  This function should identify the message type and, */
  /* if pertinent, process the arguments associated with it.  The */
  /* function expects to be called with the following parameters: */
  /*    (*notify_fn)(state_t *state, window_t *send_win, */
  /*                    window_t *recv_win, int message, ...) */
  int (*notify_fn)();
  
  /* event_fn is the function to be invoked to handle special events */
  /* encountered by the event handler.  These are generally window- */
  /* specifiec events such as manipulating a colormap when certain */
  /* mouse buttons are held down while moving the mouse.  The */
  /* function is called with the following parameters: */
  /*    (*event_fn)(state_t *state, window_t *window, */
  /*                long dev_id, short data) */
  /* where dev_id and data are the device identifier of the event and */
  /* data is the data value associated with it. */
  int (*event_fn)();
  
  /* fastdraw_fn is the function use to redraw while dragging with */
  /* the mouse.  The function is substituted when redrawing inside */
  /* mouse-driven movement routines in order to speed redrawing with */
  /* a simpler representation.  When the pointer is NULL, the normal */
  /* drawing routines are used.  The function is called with the */
  /* following parameters: */
  /*    (*fastdraw_fn)(state_t *state, window_t *window) */
  int (*fastdraw_fn)();
  
  int use_fastdraw;
  
  /* destroy_fn is the function to be invoked when destorying this */
  /* window.  This function frees associated data structures, cleanly */
  /* breaks linked lists, and leaves the window record itself to the */
  /* calling routine. The function is called with the following */
  /* parameters: */
  /*    (*destroy_fn)(state_t *state, window_t *window); */
  int (*destroy_fn)();
  
  /* WINDOW LINKED LIST: */
  /* next points to the next window record.  This is primarily used */
  /* for initializing variables in the records for process_events. */
  struct window_s *next;
};


/* -------------------------------- status -------------------------------- */
/* Contains all of the old GUImainWindow fields. */
typedef struct status_s 
{
  XmxMenuRecord *menubar;
  Widget load_script;
  
  Widget command;
  Widget command_list;
  Widget status;
} status_t;

/* ------------------------------- animate -------------------------------- */
typedef struct animate_s
{
  /* Scale. */
  Widget frame;
  
  /* Buttons. */
  Widget forward, forward_step, stop, reverse_step, reverse;
  /* Text for frame number. */
  Widget frametext;
  
  /* Scales. */
  Widget delay, skip;
  
  /* Labels. */
  Widget frame_label, delay_label, skip_label;
  
  /* Flag to indicate whether control elements are insensitive. */
  int blank;
} animate_t;


/* -------------------------------- xform --------------------------------- */
typedef struct xform_s 
{
  /* used is TRUE if this transformation has any affect. */
  int used;
  
  /* coord is an array of transformation parameters for this */
  /* particular transformation. */
  float coord[DIMS];
} xform_t;

/* ------------------------------- longused ------------------------------- */
typedef struct longused_s 
{
  /* used is TRUE if this value has any affect. */
  int used;
  
  /* Flag values accociated with the record. */
  int value;
} longused_t;


/* ------------------------------- objinfo -------------------------------- */
struct objinfo_s 
{
  /* refs is a count of the number of objects making references to this */
  /* record.  If refs is zero, nothing is referencing this object.  The */
  /* value is incremented every time this objinfo is added to a */
  /* structure. */
  int refs;
  
  /* picked is an array of flags indicating which vertices, edges, or */
  /* faces are selected. */
  char *picked;
  
  /* size is the maximum number of items that can fit in picked. */
  long size;
  
  /* pickmode is the type of data that was selected */
  /* (PICK_NONE,PICK_VERTEX, PICK_EDGE, PICK_FACE, */
  /* PICK_OBJECT).  If pickmode != PICK_NONE, then */
  /* something may be selected, depending on the */
  /* value of hist.  If pickmode == PICK_OBJECT, */
  /* then regardless of the value of hits, this */
  /* object is selected. */
  pickmode_t pickmode;
  
  /* hits is the number of selected items referenced */
  /* by this or subordinate data sets. */
  long hits;
  
  /* first_element_id is the sequence number of the first selected */
  /* element in this object. */
  long first_element_id;
  
  /* max_hits is the number of items in this or subordinate */
  /* data sets. */
  long max_hits;
  
  /* TRANSFORMATIONS */
  /* rot contains rotation transformation information.  THETA and PHI */
  /* values describe rotation in degrees of the object.  If used is */
  /* FALSE, rot is assumed to be no change. */
  xform_t rot;
  
  /* scl contains X, Y, and Z scaling values for the object.  If used */
  /* is FALSE, scl is assumed to be no change. */
  xform_t scl;
  
  /* trn is the translation transformation to be applied to the */
  /* object.  If used is FALSE, then trn is assumed to be no change. */
  xform_t trn;
  
  /* ATTRIBUTES */
  /* draw_fn is a pointer to the drawing function used to render this */
  /* object.  The drawing function knows how to use the vdatas */
  /* associated with the sub-objects.  It is called as: */
  /*    (*draw_fn)(state_t *state, window_t *window,          */
  /*               object_t *obj, objinfo_t *objinfo)         */
  /* If the pointer is NULL, then the previously active drawing */
  /* function remains active. */
  int (*draw_fn)();
  
  /* scatterdraw_fn is a pointer to the drawing function used to  */
  /* render this object on scatter planes.  It is called as: */
  /*    (*scatterdraw_fn)(state_t *state, window_t *window,  */
  /*               object_t *obj, objinfo_t *objinfo)        */
  int (*scatterdraw_fn)();
  
  /* Rendering attributes for this object.  The interpretation of these */
  /* fields varies depending on the type of object.  At this point */
  /* these are only flags for rendering style (POINTS, LINES, */
  /* POLYGONS, VOLRENDER, ISOSURFACE, etc), shading model (FLAT, */
  /* GOURAUD, LIGHTED, etc.), overlays (AXES, BOX, OUTLINES, LABELS) */
  /* and any others that we come up with. */
  /* Also need information about color mappings, and so on... */
  /* ATTRIBUTES (according to GL):
     ___________________________________
     |___________Attributes____________|
     |backbuffer         linewidth     |
     |cmode or RGBmode   lsrepeat      |
     |color              pattern       |
     |concave            font          |
     |drawmode           RGB color     |
     |frontbuffer        RGB writemask |
     |linestyle          shademodel    |
     |                   writemask     |
     |_________________________________|
     */
  /* hereonly describes general attributes which are applied to this */
  /* level only.  These include:  ATPOINT, BOXWIRE, BOXSHADED, and */
  /* LABELS. */
  longused_t hereonly;
  
  /* heredown describes general attributes which are applied to this */
  /* level and those below.  These include:  OUTLINES, CULL, CONTOUR. */
  longused_t heredown;
  
  /* vertex, polygon, and element describe how these parts are to be */
  /* displayed.  vertex values include:  DOTS/SPHERES/JACKS.  Polygon */
  /* values include WIRE/SOLID/TRANSP and SHRINK.  Element values */
  /* include WIRE/SOLID/TRANSP and SHRINK. */
  longused_t vertex;
  longused_t polygon;
  longused_t element;
};


/* -------------------------------- object -------------------------------- */
typedef struct object_s {
  /* type is the type of subobject this is.  If type is OBJECT, then */
  /* This record refers to another group of objects.  If the type is */
  /* ATOM, then this points to a leaf object and the vdatas may be */
  /* accessed directly.  An INACTIVE bit is set if this group and its */
  /* children are "turned off". */
  int type;
  
  /* name is the name of the vdata associated with this record. */
  char *name;
  
  /* info is a pointer to the object's attributes and */
  /* transformation information.  If no special information exists */
  /* for this object, the pointer is NULL. */
  objinfo_t *info;
  
  /* refs is a count of the number of objects making references to this */
  /* object.  If refs is zero, nothing is referencing this object.  The */
  /* value is incremented every time this object is added to a */
  /* structure. */
  int refs;
  
  /* stats contains the vdata statistics for all the child objects of */
  /* this object.  Fields contain the min, max, mean, mean_sq, and */
  /* variance values for all of the child objects.  If all */
  /* of the child objects have the same values for rec_count or */
  /* rec_size, those fields contain real values.  Otherwise, they */
  /* contain -1. */
  stats_t stats[MAXDATADIMS][MAXVDATAS];
  
  union 
    {
      struct grp_s 
        {
          /* num is the number of subobjects which make up */
          /* this object. */
          int num;
          
          /* children is a list of objects which form this */
          /* object. */
          struct object_s *children;
        } grp;
      
      struct obj_s 
        {
          /* nvert is the number of vertices per polygon.  If */
          /* there are no polygons, nvert is 1. */
          int nvert;
          
          /* vdata is an array of pointers to the vdatas that */
          /* make up this object. */
          Vdata_t *vdata[MAXVDATAS];
        } obj;
    } u;
  
  /* sibling points to the next object in this level */
  /* of the hierarchy. */
  struct object_s *sibling;
} object_t;


/* -------------------------------- frame --------------------------------- */
struct frame_s 
{
  /* name is the unique descriptor of this frame. */
  char name[MAXNAMELEN];
  
  /* object points to the object groups associated with this frame of */
  /* the sequence. */
  object_t *object;
  
  /* prev and next point to the previous and next frames in the */
  /* sequence. */
  struct frame_s *prev;
  struct frame_s *next;
};


/* ---------------------------- GUIloadVsetBox ---------------------------- */
/* This is the dialog box used for loading Vsets (not scripts).  It
   provides a template for constructing a sequence of commands along
   the lines of 'data file ...', 'data frame ...', 'data x-axis ...',
   all the way up to 'data load'. */
typedef struct GUIloadVsetBox
{
  /* This is the dialog itself.  'workarea' will be its single
     child (work area) and all other Widgets here will be children
     of workarea. */
  Widget base;
  
  Widget workarea;  /* form */
  Widget x_text, y_text, z_text;
  Widget connect_text, color_text;
  Widget frame_text;
  
} GUIloadVsetBox;


typedef struct GUIoutlineBox
{
  Widget base;
  
  Widget color_scale;
} GUIoutlineBox;

typedef struct GUIlinewidthBox
{
  Widget base;
  
  Widget width_scale;
} GUIlinewidthBox;

typedef struct GUIsphereBox
{
  Widget base;
  
  Widget size_text;
  Widget depth_scale;
} GUIsphereBox;



/* ------------------------------- polyview ------------------------------- */
/* polyview_t describes a polygon-viewing window. */
typedef struct polyview_s 
{
  /* probe points to the first in a series of probe windows associated */
  /* with this polyview window.  NULL if there are no probes active. */
  window_t *probe;
  
  /* active window thread points to the window structure of the next */
  /* active window.  NULL if there are no further active windows in  */
  /* the list */
  window_t *active_window_thread;
  
  /* NEW FIELDS */
  XmxMenuRecord *menubar;
  int current_select_token;
  
  /* Toggles, below menubar. */
  Widget active, dtmlockout;
  
  /* Pushbutton for image blast. */
  Widget blast;
  
  /* Drawing volume and its state.
     Maybe store origin here also. */
  Widget drawing_volume;
  int dv_height, dv_width;
  
  GUIloadVsetBox *loadVsetBox;
  
  GUIoutlineBox *outlineBox;
  GUIlinewidthBox *linewidthBox;
  GUIsphereBox *sphereBox;
  
  /* FSB for Rayshade 4.0 output file */
  Widget ray_box;
  
  /* FSB for RGB output file */
  Widget rgb_box;
  
#ifdef RENDERMAN
  /* FSB for Renderman output file. */
  Widget rm_box;
#endif
  
  /* FSB for HDF image output. */
  Widget hdf_box;
  
} polyview_t;


/* -------------------------------- action -------------------------------- */
typedef struct action_s 
{
  /* action fn is the function to be executed when this action is */
  /* executed.  The function expects to be called with the following */
  /* parameters: */
  /*    (*action_fn)(state_t *state, action_t *action) */
  int (*action_fn)();
  
  /* count is how many times this action would like to have to */
  /* properly step through its operations. */
  int count;
  
  /* Countdown indicates how many times this action must be */
  /* executed before it can leave the queue.  If the value is less */
  /* than or equal to one before evaluation and execution, this is */
  /* its last time. */
  int countdown;
  
  /* global is TRUE if the command is independent of the active */
  /* windows.  It is FALSE otherwise. */
  int global;
  
  /* argc is the number of arguments in the action string sent into */
  /* the program.  */
  int argc;
  
  /* arg is an array containing the descriptions of, types of, and */
  /* pointers to the arguments of the action.  */
  arg_t arg[MAXACTIONARGS];
  
  struct action_s *next_serial;
  
} action_t;


/* -------------------------------- state --------------------------------- */
/* Current state of the system */
typedef struct state_s 
{
  /* active views are a list of windows that have been activated for */
  /* control purposes.  Whenever an action is taken it affects all of */
  /* the windows in the active windows list.  This pointer is NULL */
  /* when no windows are active.  If no windows are active and a */
  /* becomes the current window it also becomes the active window. */
  window_t *active_windows;
  
  /* current window is a pointer to _the_ window that the cursor was */
  /* last detected in.  When the cursor is outside of the program's */
  /* windows, this pointer is NULL.  This window's menus are used */
  /* whenever the mouse's menu button is depressed. */
  /* THIS IS KINDA USELESS NOW, SI? */
  window_t *current_window;
  
  /* palette window is a pointer to the window displaying the active */
  /* colormap and associated data values.  Whenever the cursor enters */
  /* a new current window owned by the program, the palette window is */
  /* updated with that window's data. */
  window_t *palette_window;
  
  /* animate window */
  /* This is the time window, kind of.  Hahahaha. */
  window_t *animate_window;
  
  /* status is a pointer to the status window.  If the window has not */
  /* been opened the value is NULL. */
  /* Why is this not named status_window?  Hell if I know.  Blame
     Calvert, not me.  I'm not gonna change it now. */
  window_t *status;
  
  /* info_window is a pointer to the info window.  Right now we only
     allow one of these to be open at once; tough shit to anyone who
     wants it different; give me a raise. */
  window_t *info_window;
  
  /* windows is a pointer to the list of windows the program owns.  If */
  /* no windows are open, the pointer is NULL */
  window_t *windows;
  
  
  
  /* event count is incremented by one each time a new event is */
  /* is received.  This value is used to determine the age of datasets */
  /* when swapping out data to make more room. */
  long event_count;
  
  /* redraw count is incremented by one each time the windows are all */
  /* redrawn.  This is used to determine when animations should be */
  /* advanced and when redraws are really necessary. */
  long redraw_count;
  
  /* auto_geometry is a flag which indicates whether automatic */
  /* placement and sizing of windows is to be used.  If MANUAL, */
  /* the user is allowed to manually place all windows.  If AUTO, */
  /* windows are always automatically placed.  If any other value, */
  /* (PREFMAN or PREFAUTO) the present window has had a preferred size */
  /* specified, but the next window will return to the previous state. */
  int geometry;
  
  /* nextx and nexty describe the coordinate of the lower left corner */
  /* of the next window to be posted.  These values are incremented by */
  /* incx and incy every time a window is opened, if auto_geometry is */
  /* true. */
  long nextx, nexty;
  
  /* incx and incy are the values used to increment the position of 
     the lower left corner of the next window by. */
  long incx, incy;
  
  /* sizex and sizey are the default size of the windows. */
  long sizex, sizey;
  
  
  /* script fp is the file pointer to the script file, if it is open. */
  /* Otherwise it is NULL. */
  FILE *script_fp;
  
  /* transcript fp is the file pointer to the transcript file, if it */
  /* is open.  Otherwise it is NULL. */
  FILE *transcript_fp;
  
  
  /* parse table is a pointer to the current structure used to parse */
  /* commands from the script file or stdin. */
  parse_t *parse_table;
  
  /* Current flag controls where actions are added to the action list. */
  /* If current is TRUE, the present command is defined as the command */
  /* before the currently executing command, and if not, then it is */
  /* defined as the last command in the action list. */
  int current;
  
  
  /* action head is a pointer to the next action to be executed from */
  /* the action list.  It typically points to the head of the list. */
  /* If no actions are pending, action_head is NULL. */
  action_t *action_head;
  
  /* action tail points to the last action in the action list.  All */
  /* new, non-immediate actions are added to follow the record */
  /* pointed to by this pointer.  If this pointer is NULL, new */
  /* actions get pointed to by action_head and action_tail. */
  action_t *action_tail;
  
  /* action free is a list of action records that were discarded after */
  /* use and await reuse.  If there are no free action records, the */
  /* pointer is NULL. */
  action_t *action_free;
  
  
  /* frames is a pointer to the list of frames that have been created */
  /* by the program. */
  frame_t *frames;
  
  /* files is a pointer to the tree of Vset data loaded by the 
     program. */
  files_t *files;
  
  /* ---------------------------------------------------------------------- */
  /* NEW X STUFF */
  /* base always points to the single status window --- i.e. the
     toplevel Widget in the big hierarchy. */
  Widget base;
  
  /* app_context is same for all windows */
  XtAppContext app_context;
  
  /* flag to indicate whether or not colormaps for GL windows
     have been installed */
  int gl_initialized;
  
  /* Dialog box to confirm exit from program. */
  Widget exitbox;
  
  /* Dialog box for loading a script. */
  Widget scriptbox;
  
  /* Flag to freeze actions from occurring globally. */
  int no_actions;
  
  /* Flag to indicate to parsing routines _not_ to add last
     command to GUI history list. */
  int last_cmd_already_added;
  
  /* Flag to indicate whether next Polyview window should automatically
     be placed under DTM control. */
  int next_window_dtm;
  
#ifdef DO_AUDIO
  /* SOUND support */
  /* can_use_sound is 1 if sound hardware is available;
     use_sound is 1 if we have sound turned on within the
     application. */
  int can_use_sound, use_sound;
#endif
  
  /* Global palette struct. */
  short pal[256][3];
  
  /* Zbuffer stuff. */
  
  /* zmax, zmin are ranges for the Zbuffer, to be fed to lsetdepth. */
  long zmax, zmin;
  
  /* zmax_comp, zmin_comp are ranges for a compressed Zbuffer. */
  long zmax_comp, zmin_comp;
  
} state_t;


#endif /* not __POLYVIEW_PVDATA_H__ */
