/*****************************************************************************
 * NCSA Polyview 3.1                                                         *
 *                                                                           *
 * Version 3.1 changes and additions by Gilles Bourhis.                      *
 * Version 3 changes and additions by Marc Andreessen.                       *
 * Version 2 by Brian Calvert.                                               *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvdtm.c,v 1.3 93/08/13 13:55:18 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvdtm.c,v $
 * Revision 1.3  93/08/13  13:55:18  gbourhis
 * changes in COMreceive_vdata() and _send_vdatas_in_vgroup().
 * 
 * Revision 1.2  93/07/13  16:35:46  gbourhis
 * add COMreceive_srv() to handle connect SRV message.
 * 
 * Revision 1.1  92/09/18  10:55:26  marca
 * Initial revision
 * 
#endif

#include "pv.h"
#include <net.h>
#include <netdata.h>
#include <dtm.h>
#include <srv.h>
#include <sdl.h>

/*
  This file implements support for Vsets via DTM.  DTM will only deal
  with Vdata's; each Vdata will come with a ``magic path'' which will
  uniquely identify it within a Vset (aka Vgroup tree).  What makes the
  a Vdata's path ``magic'' is the fact that it implicitly specifies the
  structure of the Vset (which is never sent, per se, via DTM); thus,
  when we receive a new Vdata, we create whatever Vgroups are necessary
  to place the new Vdata in its proper position.
  
  Calling Interfaces:
  
  External routines have the prefix COM, for communication.
  
  COMsend_vset
  COMdtm_in
  COMdtm_out
  COMpoll
  
  To Do:
  
  Data sharing is greatly lacking.  We should be considering each Vset
  to be a DAG, not a tree.
*/

/* Creation of an input port implies done_init and done_register
   as well as done_inport.  Creation of an output port implies 
   all of these (since an input port is always created prior to
   creating an output port). */

/* Note that this design specifically excludes the possibility
   of either multiple input or output ports, even in separate windows
   of the application.  This bears thought. */

static int done_init = 0;       /* called NetInit?? */
static int done_register = 0;   /* called NetRegisterModule?? */
static int done_outport = 0;    /* called NetCreateOutPort?? */
static int done_inport = 0;     /* called NetCreateInPort?? */


/* ------------------------------------------------------------------------ */
/* --------------------------- UTILITY ROUTINES --------------------------- */
/* ------------------------------------------------------------------------ */

static void _dump_magic_path (VdataPathElement *magic_path, int depth)
{
  int i;

  fprintf (stdout, "  [_dump_magic_path] depth is %d\n", depth);
  for (i = 0; i < (depth + 1); i++)
    fprintf (stdout, "  [_dump_magic_path] %d> %d %s\n", i, 
             magic_path[i].nodeID,
             magic_path[i].nodeName);
  fprintf (stdout, "  [_dump_magic_path] ************ DONE\n");
  fflush (stdout);

  return;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------- RECEIVE -------------------------------- */
/* ------------------------------------------------------------------------ */

/* This routine is the callback for receiving a new vdata
   from the NET module. */
static void COMreceive_vdata (Data *data, caddr_t state_addr)
{
  /* The header vgroup for this vset. */
  Vgroup_t *vset;
  
  /* The vgroup we're currently working on, and its child. */
  Vgroup_t *vptr, *vptr_child;

  /* Depth in the path so far. */
  int depth;

  /* The new vdata being received. */
  Vdata_t *vdata;
  int vdata_id       = data->nodeID;
  int vdata_type;                        /* This is NOT data->dost. */
  char *vdata_name   = data->nodeName;
  char *vdata_fields = data->fields;
  char *vdata_data   = data->data;

  /* The magic path for this vdata. */
  VdataPathElement *magic_path = *(data->magicPath);
  int path_length              = data->pathLength;

  /* The state and window structs. */
  state_t *state = (state_t *)state_addr;
  window_t *win;  /* Grab the window out of state's list. */

  /* Reset flag. */
  int reset;

  /* ---------------------------------------------------------------------- */
  
  /* Sanity-check the net module. */
  if (data->dot != DOT_VData)
    {
      fprintf (stderr, 
               "[COMreceive_vdata] Object other than Vdata rejected.\n");
      fflush (stderr);
      return;
    }
  if (data->dost != DOST_Float && data->dost != DOST_Int32)
    {
      fprintf 
        (stderr, 
         "[COMreceive_vdata] Polyview only supports floats and ints!\n");
      fflush (stderr);
      return;
    }

  /* We should probably support doubles as well, and convert on
     the fly. */

  /* ---------------------------------------------------------------------- */

  /* Figure out which window we're dealing with. */
  win = NULL;
  FOR_ALL_WINDOWS(state, win)
    {
      files_t *file;
      file = WIN_SOURCE(win);
      if ((file != NULL) && (file->type == DTMPORT))
        break;
    }

  /* Can't find a window to receive data into. */
  if (win == NULL)
    {
      if ((win = GetActiveWindow(state)) == NULL)
	return;
      data_dtm_explicit(state, win, "dtm");
    }
  else
    win_activate(state, win);
      
#ifdef PRIVATE_SUPPORT
  /* Check for privacy enabled in that window. */
  if (win->private)
    return;
#endif

  stprintf (state, "Receiving vdata %s via DTM...\n", vdata_name);  

  /* The first element of the magic path identifies the vset 
     to which this vdata belongs.  Go try to find it. */
  vptr = find_vgroup (state, WIN_SOURCE(win), magic_path[0].nodeName);

  /* If it exists, we now have a pointer to it.  If not,
     create a new vset. */
  reset = FALSE;
  if (vptr == NULL)
    {
      vptr = add_vgroup (state, magic_path[0].nodeID, magic_path[0].nodeName,
                         WIN_SOURCE(win)->groups);
      vptr->children_read = TRUE;
      reset = TRUE;
    }

  /* Keep this vset around for reference. */
  vset = vptr;

  /* ---------------------------------------------------------------------- */

  /* path_length is the number of elements in the array magic_path.
     In other words, magic_path is defined for [0..path_length-1]. */
  
  /* depth is how deep we are in the path so far.  Since we've already
     dealt with the first element of magic_path, depth = 1 here. */
  depth = 1;
  
  /* We will have reached the parent vgroup when we have found
     the vgroup corresponding to magic_path[path_length-1].
     As long as we have not reached that point, keep iterating. */
  while (depth < (path_length - 1))
    {
      /* Try to find the child of vptr with this name/id. */
      vptr_child = GetVgroupInVgroup (state, vptr, magic_path[depth].nodeName);

      /* If child doesn't exist, make a child with vptr as parent. */
      if (vptr_child == NULL)
        {
          vptr_child = add_vgroup
            (state, magic_path[depth].nodeID, magic_path[depth].nodeName, 
             vptr);
          vptr_child->children_read = TRUE;
        }
      
      /* Make the child the new parent. */
      vptr = vptr_child;
      /* Go a level deeper. */
      depth++;
    }

  /* ---------------------------------------------------------------------- */

  /* When we hit this point, we assume vptr is the parent vgroup for
     this vdata. */

  /* Figure out what type of data we're dealing with. */
  switch (data->dost)
    {
    case DOST_Float:
      vdata_type = PVFLOAT;
      break;
    case DOST_Int32:
      vdata_type = PVINTEGER;
      break;
    default:
      stprintf (state, "[COMreceive_vdata] ERROR: Unknown type %d.\n",
                data->dost);
      break;
    }
  
  /* Go find the vdata in vptr. */
  vdata = get_vdata_in_vgroup (state, vptr, vdata_name);

  if (vdata == NULL)
    {
      /* If a vdata didn't already exist, make one. */
      vdata = add_vdata (state, vdata_id, vdata_name, vptr);

      /* Set the vdata parts. */
      /* vdata->fields is an array */
      strcpy (vdata->fields, vdata_fields);
      vdata->stats_read = FALSE;
      vdata->data = vdata_data;
      vdata->in_memory = TRUE;

      /* Set the stats parts. */
      vdata->stats[0].type = vdata_type;
      /* rank is always 2, dim[0] = number of elements per record
         and dim[1] = number of records */
      vdata->stats[0].rec_size = data->dim[0];
      vdata->stats[0].rec_count = data->dim[1];

      calc_stats (state, vdata);
    }
  else
    {
      /* If the vdata already exists (less likely), then just
         replace its data. */
      /* replace_vdata_contents (state, vdata_data, ...); */
      /* For now, punt. */
      fprintf (stdout, "[COMreceive_vdata] Vdata already exists!\n");
      fflush (stdout);
    }

  stprintf (state, "Receiving vdata %s via DTM... done\n", vdata_name);

#if 0
  if (vdata_type == PVINTEGER)
    {
      int i;
      PRT (("FOR VDATA %s\n", vdata_name));
      PRT (("  size %d count %d\n", 
            vdata->stats[0].rec_size, vdata->stats[0].rec_count));
      for (i = 0; 
           i < vdata->stats[0].rec_size * vdata->stats[0].rec_count; i++)
        {
          PRT (("%d ", ((int *)(vdata->data))[i]));
        }
    }
#endif
  
  /* Do resets of time, info, whatever. */
  /* This is probably a bad bad idea for each incoming vdata... */
  WIN_GROUP(win) = set_vgroup (state, WIN_SOURCE(win), WIN_GROUPNAME(win));

  /* That's all, folks. */
  return;
}


/* ------------------------------------------------------------------------ */
/* --------------------------------- SEND --------------------------------- */
/* ------------------------------------------------------------------------ */

/* ------------------------ _send_vdatas_in_vgroup ------------------------ */
static void  _send_vdatas_in_vgroup (Vgroup_t *vptr, 
                                     VdataPathElement *magic_path,
                                     int depth)
{
  /* We receive a vgroup, the current magic path, and depth, which
     is the depth of the vgroup in the vset.  The length of the magic
     path is thus depth + 1. */

  /* Vdata pointer. */
  Vdata_t *vdptr;

  /* Type to send, DOST_Float or DOST_Int32. */
  int type;

  /* Return code from send. */
  int rc;

  if (!done_outport)
    {
      stprintf (gstate, "ERROR trying to send with no outport\n");
      return;
    }

  /* We start with the first child. */
  vdptr = vptr->vdatas;
  
  /* In this vdata:
     We have array stats[MAXDATADIMS] == stats[4].
     MAXDATADIMS = maximum width of a vdata.
     Each stats has:
       type (PVINTEGER or PVFLOAT)
       rec_count (number of records)
       rec_size (number of elements per record).
     Refer to stats[0] only.  That's just the way it works. */
  
  /* While we have child vdatas left to send, send 'em. */
  for (; vdptr != NULL; vdptr = vdptr->sibling)
    {
      if (! vdptr->in_memory)	/* skip it if not in memory */
	continue;

      switch (vdptr->stats[0].type)
        {
        case PVINTEGER:
          type = DOST_Int32;
          break;
        case PVFLOAT:
          type = DOST_Float;
          break;
        default:
          fprintf (stderr, 
                   "[_send_vdatas_in_vgroup] OOPS Polyview type of %d\n",
                   vdptr->stats[0].type);
          fflush (stderr);
          return;
        }
      
      rc =
        NetSendVData
          (NULL,                        /* netport, or NULL for all */
           vdptr->name,                 /* label, duplicate of nodeName */
           &magic_path,                 /* magic path */
           depth + 1,                   /* magic path length */
           vdptr->id,                   /* node ID */
           vdptr->name,                 /* node name */
           " ", /* vdptr->fields, */    /* fields in vdata */
           vdptr->stats[0].rec_count,   /* number of records */
           vdptr->stats[0].rec_size,    /* number of elements */
           type,                        /* type */
           vdptr->data,                 /* the data itself */
           FALSE,                       /* copy before returning */
           FALSE,                       /* distribute internally */
           NULL                         /* if ^^ is TRUE, this is our name */
           );

    }
  
  /* That's all, folks. */
  return;
}

/* --------------------------- _traverse_vgroup --------------------------- */

static void _traverse_vgroup (Vgroup_t *vptr, VdataPathElement *magic_path,
                              int depth)
{
  /* We receive an arbitrary vgroup, a partially constructed magic_path,
     and an arbitrary depth in a vset. */
  char *name;
  
  /* Load this element of the magic path. */
  name = (char *)PVMALLOC (strlen (vptr->name) + 1);
  if (name == NULL)
    {
      fprintf (stderr, "YOW! PVMALLOC NULL in _traverse_vgroup.\n");
      fflush (stderr);
    }
  
  magic_path[depth].nodeName = 
    strcpy ((char *)PVMALLOC (strlen (vptr->name) + 1), vptr->name);
  magic_path[depth].nodeID = vptr->id;
  
  /* Check for vdatas. */
  if (vptr->vdatas != NULL)
    _send_vdatas_in_vgroup (vptr, magic_path, depth);
  
  /* Check for child vgroup; if there's more than one, the sibling
     pointer will handle the rest. */
  if (vptr->groups != NULL)
    _traverse_vgroup (vptr->groups, magic_path, depth + 1);
  
  /* Check for sibling vgroup. */
  if (vptr->sibling != NULL)
    _traverse_vgroup (vptr->sibling, magic_path, depth);
  
  /* That's all, folks. */
  return;
}


/* Does this make sense? */
#define MAX_MAGIC_PATH_LENGTH 20

/* ----------------------------- COMsend_vset ----------------------------- */

/* This routine is called from Polyview and broadcasts an entire
   Vset.  This is all Polyview will probably do by itself, since it
   doesn't actually create data. */
void COMsend_vset (state_t *state, Vgroup_t *vset)
{
  /* The current vgroup. */
  Vgroup_t *vptr;
  
  /* The current path. */
  VdataPathElement *magic_path;
  
  /* The current depth in the path.  0 means we're at the base. */
  int depth;
  
  if (!done_outport)
    {
      fprintf (stderr, "ERROR trying to send vset with no outport\n");
      fflush (stderr);
      return;
    }
  
  /* We want to do a depth-first traversal of vset and
     its children.  Currently we pretend we're dealing with
     a tree, even though it's really a DAG.  Tough shit. */
  
  /* A traversal involves tracking the current location
     in the tree via magic_path. */
  magic_path = 
    (VdataPathElement *)PVMALLOC 
      (MAX_MAGIC_PATH_LENGTH * sizeof (VdataPathElement));
  
  /* We're currently pointing at the head of the tree. */
  depth = 0;
  vptr = vset;
  
  /* Traverse the tree.  At any point in here, magic_path_length
     will be depth + 1; this is important. */
  _traverse_vgroup (vptr, magic_path, depth);
  
  /* That's all, folks. */
  return;
}


/* ------------------------------------------------------------------------ */
/* ---------------------------- COMreceive_sdl ---------------------------- */
/* ------------------------------------------------------------------------ */

#define MAX_VERTICES 10000

/* Accept three floating-point values and build an index into the
   Polyview RGB colormap. */
#define	RGB_TO_CMAP(r,g,b) \
  ((((int)(b*3.0))<<6) | (((int)(g*7.0))<<3) | (((int)(r*7.0))))
#define	PVREALLOC(p,n,t,r) \
  ((p) ? (t (*) r)realloc(p,n*sizeof(t r)) : (t (*) r)calloc(n, sizeof(t r)))

/* Given a vgroup, read vertices into vdatas named "pcoord"
   and "color". */
static int ReadVertexList (state_t *state, Vgroup_t *vg, Data *data)
{
  int	i, vcount;
  float	r, g, b;
  float	mag;
  int	*colorp;
  float	(*pcoord)[DIMS];
  int	size, count;
  Vdata_t *color_v, *pcoord_v;
  struct DTM_TRIPLET *primbuf;
  
  /* Find the existing vdatas in the group, if possible. */
  /* If they don't exist, create them. */
  if ((pcoord_v = get_vdata_in_vgroup(state, vg, "pcoord")) == NULL) 
    {
      pcoord_v = add_vdata(state, -1, "pcoord", vg);
      pcoord_v->data = NULL;
      pcoord_v->in_memory = TRUE;
      size = pcoord_v->stats[0].rec_count = 0;
      pcoord_v->stats[0].rec_size = 3;
      pcoord_v->stats[0].type = PVFLOAT;
    }
  else 
    size = pcoord_v->stats[0].rec_count;
  
  if ((color_v = get_vdata_in_vgroup(state, vg, "color")) == NULL) 
    {
      color_v = add_vdata(state, -1, "color", vg);
      color_v->data = NULL;
      color_v->in_memory = TRUE;
      color_v->stats[0].rec_count = 0;
      color_v->stats[0].rec_size = 1;
      color_v->stats[0].type = PVINTEGER;
    }
  
  if (pcoord_v->stats[0].rec_count != color_v->stats[0].rec_count) 
    {
      stprintf(gstate, "SYSTEM ERROR: pcoord and color sizes don't match!\n");
      return ST_ERROR;
    }
    
  /* Allocate space for the color and pcoord data. */
  colorp = (int *) color_v->data;
  pcoord = (float (*)[DIMS]) pcoord_v->data;
  count = 0;
    
  /* Now data->dim[1] = number of DTM_TRIPLETS.  This will be vcount. */
  vcount = data->dim[1];

  /* primbuf is created now as a cast from the character data we
     receive from the net module. */
  primbuf = (struct DTM_TRIPLET *)(data->data);
  
  /* Initially size is 0 (meaning we get a realloc first thing).
     i will range from 0 to the number of primitives (coords
     plus colors) in the dataset. */

  /* For each triplet, read it.  For each position, increment count
     of coords.  When we max out the current allocation of coords
     (three floats and one int in pcoord and colorp respectively)
     allocate more space. */
  for (i = 0; i < vcount; i++)
    {
      if (count == size) 
        {
          /* We've maxed out current storage; go get more. */
          size += MAX_VERTICES;
          colorp = PVREALLOC(colorp, size, int, );
          pcoord = PVREALLOC(pcoord, size, float, [DIMS]);
        }
      
      if (primbuf[i].tag == SDLposition)  
        {
          /* Add the current color and position to their vdatas
             and increment the count of coords. */
          colorp[count] = RGB_TO_CMAP(r, g, b);
          pcoord[count][X] = primbuf[i].x;
          pcoord[count][Y] = primbuf[i].y;
          pcoord[count][Z] = primbuf[i].z;

          count++;
        }
      
      else if (primbuf[i].tag == SDLcolor)  
        {
          /* All we do is store this color for use with the next
             position. */
          r = primbuf[i].x;
          g = primbuf[i].y;
          b = primbuf[i].z;
        }
      
      else if (primbuf[i].tag == SDLnormal) 
        {
          /* Old normal code taken out... Polyview doesn't do
             normals yet, but it will derive them on a polygon
             basis in the drawing routines. */
        }
    }

  /* Trim off the extra memory from the arrays. */
  colorp = PVREALLOC(colorp, count, int, );
  pcoord = PVREALLOC(pcoord, count, float, [DIMS]);
  
  /* Store the data as vdatas, deleting any existing data. */
  /* Update any instances of the vdata in the diplay tree. */
  color_v->data = (char *)colorp;
  pcoord_v->data = (char *)pcoord;
  pcoord_v->stats[0].rec_count = color_v->stats[0].rec_count = count;

  calc_stats(state, pcoord_v);
  calc_stats(state, color_v);
  
  return count;
}


/* Given a vgroup, width of each connectivity spec (3=tri, etc),
   and number of vertices, build a connectivity vdata in this
   vgroup. */
static void BuildConnectList (state_t *state, Vgroup_t *vg, 
                              int width, int vertices)
{
  Vdata_t *connect_v;
  int *connect = NULL;
  int *c;
  int i;
  char plist[MAXNAMELEN];
  int length;
  int count;

  /* Calculate the number of polygons in the list. */
  length = vertices / width;
  
  /* Get any existing connectivity data.  Otherwise, create it. */
  sprintf(plist, "plist%d", width);
  if ((connect_v = get_vdata_in_vgroup(state, vg, "plist*")) == NULL) 
    {
      connect_v = add_vdata(state, -1, plist, vg);
      connect_v->data = NULL;
      connect_v->in_memory = TRUE;
      connect_v->stats[0].rec_count = 0;
      connect_v->stats[0].rec_size = 0;
      connect_v->stats[0].type = PVINTEGER;
    }
  count = length * width;

  if ((length != connect_v->stats[0].rec_count) ||
      (width != connect_v->stats[0].rec_size)) 
    {
      connect_v->data = (char *) PVREALLOC(connect, count, int, );
      strcpy(VDA_NAME(connect_v), plist);
    }
  connect_v->stats[0].rec_count = length;
  connect_v->stats[0].rec_size = width;
  connect = (int *) connect_v->data;
  
  /* Fill the connectivity list with the appropriate connections. */
  for (i = 1, c = connect; i <= count; c++, i++)
    *c = i;
  
  /* Store the list as a vdata, deleting any existing data. */
  /* Update any instances of the vdata in the diplay tree. */
  calc_stats(state, connect_v);
  
  return;
}


static void COMreceive_sdl (Data *data, caddr_t state_addr)
{
  state_t      *state = (state_t *)state_addr;
  window_t     *win;
  char		header[DTM_MAX_HEADER];
  char		title[MAXNAMELEN];
  char	       *path;
  SDLprim_t	prim;
  int		size, count;
  Vgroup_t     *vg;
  int		reset;

  /* Figure out which window we're dealing with. */
  win = NULL;
  FOR_ALL_WINDOWS(state, win)
    {
      files_t *file;
      file = WIN_SOURCE(win);
      if ((file != NULL) && (file->type == DTMPORT))
        break;
    }

  /* Can't find a window to receive data into. */
  if (win == NULL)
    return;

  /* Determine the type of primitive:  points, lines, 
     polygons (n-gons) */
  prim = data->dost;
  strcpy (title, data->label);

  path = title;

  if (path[0] == '\0')
    /* No title associated with this dataset. */
    /* Default to '/frame0001' */
    strcpy(title, "frame0001");
  else 
    /* Skip the leading slash, if any. */
    if (*path == '/')
      path++;
  
  /* Get a pointer to the vgroup structure that should contain the */
  /* data.  (If the structure does not already exist, create it). */
  /* The vgroup should be the first that matches the name "title" */
  /* within the window's file. */
  if ((vg = set_vgroup(state, WIN_SOURCE(win), path)) == NULL) 
    {
      /* The vgroup does not exist.  Create it. */
      vg = add_vgroup(state, -1, path, WIN_SOURCE(win)->groups);
      reset = TRUE;
    }
  else
    reset = FALSE;
  
  count = ReadVertexList(state, vg, data);
  switch(prim)  
    {
    case SDLpoint:
    case SDLsphere:
      /* Do nothing... we have the vertices. */
      break;
      
    case SDLlineseg:
      BuildConnectList(state, vg, 2, count);
      break;
      
    case SDLtriangle:
      BuildConnectList(state, vg, 3, count);
      break;
      
    case SDLquad:
      BuildConnectList(state, vg, 4, count);
      break;
      
    default:
      stprintf(state, "SYSTEM ERROR: Received unsupported primitive.\n");
      break;
    }
  
  WIN_GROUP(win) = set_vgroup(state, WIN_SOURCE(win), WIN_GROUPNAME(win));
  data_load(state, win, reset);
}


/* ------------------------------------------------------------------------ */
/* ---------------------------- COMreceive_com ---------------------------- */
/* ------------------------------------------------------------------------ */

/* Oh boy, the fun part.
   We receive text commands, prefixed by "polyview". */

/* This flag prevents us from rebroadcasting events that we have 
   just received. */
static int last_command_from_dtm = 0;

static void COMreceive_com (Com *c, caddr_t state_addr)
{
  char *command;

  if (strncmp ("polyview ", c->mesg, 9) == 0)
    {
#ifdef PRIVATE_SUPPORT
      /* Check for privacy enabled. */
      if (gstate->active_windows->private)
        goto done;
#endif

      /* Oh boy, it's ours! */
      command = strcpy ((char *)PVMALLOC (strlen(c->mesg)+1), &(c->mesg[9]));

      /* Make sure we don't rebroadcast this command. */
      last_command_from_dtm = 1;

      /* Go process it.  This will result in a call to COMsend_com,
         which will be bounced right back since last_command_from_dtm
         is high. */
      parse_line ((state_t *)state_addr, command, FALSE);
    }

 done:
  return;
}


void COMsend_com (state_t *state, char *command)
{
  char *mesg;

  /* Quietly return if outport not active. */
  if (!done_outport)
    return;

  /* If the last command came from DTM, bong. */
  if (last_command_from_dtm)
    {
      last_command_from_dtm = 0;
      return;
    }

#ifdef PRIVATE_SUPPORT
  /* If we have privacy turned on, don't send anything. */
  if (state->active_windows->private)
    return;
#endif

#if 0
  /* Change moves to jumps to get where we're going quickly. */
  if (strncmp (command, "move camera", 11) == 0)
    {
      /* :-) */
      command[0] = 'j';
      command[1] = 'u';
      command[2] = 'm';
      command[3] = 'p';
    }
#endif

  /* Some commands are useless over the net. */
  if (
#ifdef PRIVATE_SUPPORT
      strncmp (command, "dtm private", 11) == 0 ||
      strncmp (command, "dtm public", 10) == 0 ||
#endif
      strncmp (command, "do ", 3) == 0 ||
      strncmp (command, "script", 6) == 0 ||
      strncmp (command, "data file", 9) == 0 ||
      strncmp (command, "data load", 9) == 0 ||
      strncmp (command, "data dtm", 8) == 0 ||
      strncmp (command, "dtm", 3) == 0 ||
      strncmp (command, "blast", 5) == 0 ||
      strncmp (command, "save", 4) == 0
      )
    return;
  
  /* This WILL NOT WORK, since we will not have loaded our own data by
     the time we reach here.  (All we've done is added our 'data load'
     command to our action list.) */
#if 0
  /* If we're loading our own data, send the data across. */
  if (strncmp (command, "data load", 9) == 0)
    {
      window_t *win = state->active_windows;
      if (WIN_GROUP(win) != NULL)
        {
          /* We assume sequential transmissions. */
          COMsend_vset (state, WIN_GROUP(win));
          NetSendCommand 
            (NULL, "Polyview", "polyview data load\0", NULL, NULL);
        }
      return;
    }
#endif
  
  mesg = (char *) PVMALLOC (strlen (command) + 20);
  sprintf (mesg, "polyview %s\0", command);
  NetSendCommand (NULL, "Polyview", mesg, NULL, NULL);

  return;
}

static void COMreceive_srv (Server *srv, caddr_t state_addr)
{
  if (srv->func == SRV_FUNC_CONNECT)
    done_outport = 1;
}

/* ------------------------------------------------------------------------ */
/* ------------------------------ COMdtm_in ------------------------------- */
/* ------------------------------------------------------------------------ */

void COMdtm_in (state_t *state, char *path)
{
  NetPort *inport;
  static char userID[200];
  
  if (!done_init)
    {
      char host[65];

      if (cuserid(userID) != NULL &&
	  gethostname(host, 64) >= 0)
	      strcat(strcat(userID, "@"), host);
      else
	      gethostname(userID, 64);

      NetInit (strcat(userID, " (Polyview 3.1)"));
      done_init = 1;
    }
  
  if (!done_register)
    {
      NetRegisterModule
        ("Polyview", NETVDATA,
         COMreceive_vdata, (caddr_t) state,
         NULL, (caddr_t) 0,
         NULL, (caddr_t) 0);
      NetRegisterModule
        ("Polyview", NETSDL,
         COMreceive_sdl, (caddr_t) state,
         NULL, (caddr_t) 0,
         NULL, (caddr_t) 0);
      NetRegisterModule
        ("Polyview", NETCOM,
         COMreceive_com, (caddr_t) state,
         NULL, (caddr_t) 0,
         NULL, (caddr_t) 0);
      NetRegisterModule
        ("Polyview", NETSRV,
         COMreceive_srv, (caddr_t) state,
         NULL, (caddr_t) 0,
         NULL, (caddr_t) 0);
      done_register = 1;
    }
  
  if (!done_inport)
    {
      inport = NetCreateInPort (path);
      /* Tell the user. */
      stprintf (state, "Created input port.\n");
      done_inport = 1;
    }
  else
    {
#if 0
      fprintf 
        (stdout, "[COMdtm_in] BONNNNNNNG quietly, already have inport\n");
      fflush (stdout);
#endif
    }
  
  return;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------ COMdtm_out ------------------------------ */
/* ------------------------------------------------------------------------ */

/* called from pvtext:do_dtm_out */
int COMdtm_out (state_t *state, window_t *win, char *port)
{
  if (!done_outport)
    {
      /* Get everything set up to support DTM. */
      /* HEY!  This looks bad... we want true two-way communication... */
      COMdtm_in (state, ":0");
      
      /* Make the output port. */
      NetCreateOutPort (port);

      done_outport = 1;
    }
  else
    {
      fprintf (stderr, "ERROR trying to make additional outport port.\n");
      fflush (stderr);
    }
  
  return ST_OKAY;
}

/* Predicate for outport active. */
int COMdtm_out_active_p (state_t *state)
{
  if (done_outport)
    return 1;
  else
    return 0;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------- COMpoll -------------------------------- */
/* ------------------------------------------------------------------------ */

void COMpoll (state_t *state)
{
  if (done_inport)
    NetClientPollAndRead ();
  
  return;
}
