/*****************************************************************************
 * NCSA Polyview 3.1                                                         *
 *                                                                           *
 * Version 3.1 changes and additions by Gilles Bourhis.                      *
 * Designed and implemented by Marc Andreessen (marca@ncsa.uiuc.edu).        *
 * (Based on Polyview 2.0 by Brian Calvert.)                                 *
 *                                                                           *
 * Software Development Group                                                *
 * National Center for Supercomputing Applications                           *
 * University of Illinois at Urbana-Champaign                                *
 *                                                                           *
 * This is BETA release software.  As such it may contain software bugs and  *
 * exhibit inconsistencies.                                                  *
 *                                                                           *
 * Please send bug reports to polyview@ncsa.uiuc.edu.                        *
 *                                                                           *
 * Copyright (c) 1992 The Board of Trustees of the University of Illinois.   *
 *                                                                           *
 * Permission to use, copy, and modify this software and its                 *
 * documentation for educational, research, and non-profit purposes is       *
 * hereby granted, provided that the above copyright notice, the original    *
 * authors names, and this permission notice appear in all such copies.      *
 * Any distribution of this software requires the explicit and written       *
 * authorization of the authors.                                             *
 *                                                                           *
 * The University of Illinois makes no representations about the             *
 * suitability of this software for any purpose.  It is provided "as is"     *
 * without warranty of any kind.                                             *
 *****************************************************************************/

/* $Id: pvpal.c,v 1.3 93/08/13 12:28:37 gbourhis Exp $ */

#ifdef RCSLOG
$Log:	pvpal.c,v $
 * Revision 1.3  93/08/13  12:28:37  gbourhis
 * Add transcript of palette fiddle, change of version number.
 * 
 * Revision 1.2  92/09/26  08:45:23  marca
 * Changed Background to Backdrop in menu.
 * 
 * Revision 1.1  1992/09/18  10:55:26  marca
 * Initial revision
 *
#endif

#include "pv.h"

#define	PALETTE_WIDTH 256
#define	PALETTESIZE 256
#define	PS_W (PALETTESIZE)
#define	PS_H 20
#define	AR_W 7
#define	AR_H 7

/* ------------------------------------------------------------------------ */
/* OK, dammit.

   Calvert overloaded window_s->image in the case of a palette window.

   What this means is that palette_window->image will change as new
   Polyview windows are made active.

   So instead of having palette_window->image hold palette window-
   specific info (e.g. important Widgets, etc.) we will store those
   local to this file. */
/* ------------------------------------------------------------------------ */

static XmxMenuRecord *pal_menubar = NULL;
static Widget pal_dv;
static int pal_dv_height, pal_dv_width;
/* This should be set to PV_NULL when the palette has been modified. */
static int current_presets_token;
/* This should always be set to something. */
static int current_background_token;
static Widget load_hdf = NULL, load_raw = NULL;
static Widget save_hdf = NULL, save_raw = NULL;

static unsigned long pal_bg = 0x0;

/* The last 'type' passed to set_palette; use this to figure out
   when to reset the menubar entries. */
static int last_set_palette = -1;



static int adjust_palette (state_t *state, palette_t *pal, float oldorigin, 
                           float oldslope, float origindel, float slopedel)
{
  float	slope_p;
    
  /* Adjust slope of mapping function */
  slope_p = pal->slope;
  pal->slope = fexp(flog(oldslope) + slopedel);
  
  if (pal->slope < 1.0/512.0) pal->slope = 1.0/512.0;
  if (pal->slope > 255.0) pal->slope = 255.0;
  
  /* Adjust old origin to account for shift */
  /* in slope */
  oldorigin -= (pal->slope - slope_p) * 128.0;
  
  /* Adjust origin of mapping function */
  pal->origin = oldorigin + origindel;
  
  /* Update the color map */
  set_palette(state, pal, PAL_MAP);
  
  return ST_OKAY;
}



/* init_palette initializes the palette window's data structures. */
/* Returns ST_OKAY. */
int init_palette(state_t *state, window_t *win, int reset)
{
  WIN_DELAY(win) = 1;
  
  return ST_OKAY;
}



/* ---------------------------- redraw_palette ---------------------------- */
int redraw_palette(state_t *state, window_t *win)
{
  palette_t *pal;
  int i;
  int xpad = (WIN_WIDTH(win)-PALETTE_WIDTH)/2;

  /* We don't have focus yet. */
  XmxWinset (pal_dv);

  /* Clear the window to the background color. */
  /* cpack (pal_bg); */
  cpack (0x0);
  clear();

  /* Want to make sure things are drawn in order given. */
  zbuffer (FALSE);
  
  /* If there is a palette associated with this window, display the */
  /* min and max values, as well as any selected color. */
  pal = (palette_t *) WIN_IMAGE(win);
  if (pal == NULL) 
    {
      /* No palette associated with this window. */
      cpack(0xffffff);
      gprintf(state, WIN_WIDTH(win)/2, WIN_HEIGHT(win)/2,
              GPR_CENTERH | GPR_CENTERV, "No active palette.");
    }
  else 
    {
      /* Draw the palette strip. */
      /* We want to draw a series of lines first,
         just to get it done.  This is kinda inefficient and
         requires that the palette stay at the fixed size. */
      
      for (i = 0; i < PALETTESIZE; i++)
        {
          bgnline ();
          PALETTECPACK (i);
          p2i (xpad+i, win->cy-PS_H);
          p2i (xpad+i, win->cy+PS_H);
          endline ();
        }
      
      /* Draw the border of the palette strip. */
      cpack (0xffffff);
      bgnclosedline();
      p2i(xpad-1, win->cy-PS_H-1);
      p2i(xpad-1, win->cy+PS_H+1);
      p2i(xpad+PS_W, win->cy+PS_H+1);
      p2i(xpad+PS_W, win->cy-PS_H-1);
      endclosedline();
      
      /* Print the minimum and maximum values on the top left */
      /* an right corners of the palette, respectively. */
      gprintf(state, xpad, win->cy+PS_H+AR_H+3,
              GPR_BOTTOMJ | GPR_LEFTJ, "%.2f", pal->min);
      gprintf(state, xpad+PS_W, win->cy+PS_H+AR_H+3,
              GPR_BOTTOMJ | GPR_RIGHTJ, "%.2f", pal->max);
      
      /* If one of the colors have been selected, display the */
      /* arrow at the top and bottom of the palette strip.  Also, */
      /* print the approximate range of scientific values. */
      if (pal->probe_value > 0) 
        {
          int i;
          
          i = pal->probe_value;
          /* Draw the bottom arrow */
          bgnpolygon();
          p2i(xpad+i, win->cy-PS_H-2);
          p2i(xpad+i+AR_W/2, win->cy-PS_H-2-AR_H);
          p2i(xpad+i-AR_W/2, win->cy-PS_H-2-AR_H);
          endpolygon();
          
          /* Print the range of scientific values that */
          /* correspond to this color. */
          if ((i > 0) && (i < PALETTESIZE) && (pal->min != pal->max)) 
            {
              float min;
              float max;
              
              min = pal->min;
              max = pal->max;
              gprintf(state, xpad+PS_W/2, win->cy-PS_H-AR_H-2,
                      GPR_TOPJ | GPR_CENTERH,
                      "(%.2f %.2f)",
                      min + (max-min)/
                      (float)(PALETTESIZE-2)*((float)i-1.5),
                      min + (max-min)/
                      (float)(PALETTESIZE-2)*((float)i-0.5));
            }
        }
    }
  
  swapbuffers();
  return ST_OKAY;
}



int animate_palette(state_t *state, window_t *win)
{
  if ((WIN_FORWARD(win) != 0) || (WIN_REVERSE(win) != 0)) 
    {
      palette_t *palette;
      
      /* Get a pointer to the palette. */
      palette = (palette_t *) WIN_IMAGE(win);
      
      adjust_palette(state, palette, palette->origin, palette->slope,
                     palette->originstep, palette->slopestep);
      
      if (palette->origin < (256.0 / palette->slope)) 
        {
          palette->origin = 256.0 / palette->slope;
          palette->originstep = -palette->originstep;
        }
      if (palette->origin > 256.0) 
        {
          palette->origin = 256.0;
          palette->originstep = -palette->originstep;
        }
            
      if (palette->slope < 1.0/512.0) 
        {
          palette->slope = 1.0/512.0;
          palette->slopestep = -palette->slopestep;
        }
      if (palette->slope > 255.0) 
        {
          palette->slope = 255.0;
          palette->slopestep = -palette->slopestep;
        }
      
      /* Update the palette */
      set_palette(state, palette, PAL_MAP);
      
      /* Go get a redraw. */
      set_redraw(state, win);
      /* process_events(state); */
      
    }

  return ST_OKAY;
}



/* ------------------------ GUIpaletteResetBackdrop ----------------------- */
#define SET_BG_TOGGLE(x) \
  XmxRSetToggleState (pal_menubar, x, (win->backdrop == x));

void GUIpaletteResetBackdrop (state_t *state, window_t *win)
{
  /* Reset palette entries in pal_menubar. */

  if (pal_menubar == NULL || win == NULL)
    return;
  
  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLACK);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_WHITE);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_GREY);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLUE);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_RED);
  
  SET_BG_TOGGLE(PV_PALETTE_BACKG_RGBY);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_RGBLUE);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_RBBLACK);
  
  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLUEHORIZ);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_REDHORIZ);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_GREENHORIZ);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_YELLOWHORIZ);
  
  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLUESKY);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_REDSKY);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_GREENSKY);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_YELLOWSKY);
  
  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLUESTREAK);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_REDSTREAK);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_GREENSTREAK);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_YELLOWSTREAK);

  SET_BG_TOGGLE(PV_PALETTE_BACKG_BLUESMEAR);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_REDSMEAR);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_GREENSMEAR);
  SET_BG_TOGGLE(PV_PALETTE_BACKG_YELLOWSMEAR);

  return;
}


void GUIpaletteResetPresets (state_t *state)
{
  if (pal_menubar == NULL)
    return;
  
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_RAINBOW,
                      (last_set_palette == PAL_RAINBOW));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_WTOB,
                      (last_set_palette == PAL_W_TO_B));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_BTOW,
                      (last_set_palette == PAL_B_TO_W));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_RGB,
                      (last_set_palette == PAL_RGB));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_YELLOW,
                      (last_set_palette == PAL_ALLYELLOW));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_REDBLUE,
                      (last_set_palette == PAL_REDBLUE));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_WHITEBLUE,
                      (last_set_palette == PAL_WHITEBLUE));
  XmxRSetToggleState (pal_menubar, PV_PALETTE_PRESETS_YELLOWRED,
                      (last_set_palette == PAL_YELLOWRED));

  return;
}




/* ------------------------------------------------------------------------ */
/* ---------------------------- EVENT HANDLERS ---------------------------- */
/* ------------------------------------------------------------------------ */

/* ------------------------- GUIpalMotion1Handler ------------------------- */
static XmxEventHandler (GUIpalMotion1Handler)
{
  int CM[DIMS];
  /* There has to be a palette window or we wouldn't be here. */
  window_t *win = gstate->palette_window;
  palette_t *pal;
  int xpad = (WIN_WIDTH(win)-PALETTE_WIDTH)/2;

  if (WIN_IMAGE(win) == NULL)
    return;

  if (event->xbutton.state == Button1Mask)
    {
      pal = (palette_t *)WIN_IMAGE(win);
      
      CM[X] = event->xmotion.x;
      CM[Y] = WIN_HEIGHT(win) - 1 - event->xmotion.y;
      
      /* If inside the window, drag the slider. */
      if ( (CM[X] >= 0) && (CM[X] < WIN_WIDTH(win)) &&
          (CM[Y] >= 0) && (CM[Y] < WIN_HEIGHT(win)) ) 
        {
          CM[X] -= xpad;
          if (CM[X] < 1)
            pal->probe_value = 1;
          else if (CM[X] >= PALETTESIZE)
            pal->probe_value = PALETTESIZE-1;
          else
            pal->probe_value = CM[X];
        }
      
      set_redraw (gstate, win);
      /* process_events (gstate); */
    }
  
  return;
}


/* ------------------------- GUIpalMotion2Handler ------------------------- */
/* File-local variables for handling button2 palette fiddling. */
static int local_x, local_y;
static float old_origin, old_slope;


static XmxEventHandler (GUIpalMotion2Handler)
{
  float slope_p;
  char line[MAXLINELEN];

  window_t *win = gstate->palette_window;

  if (event->xbutton.state == Button2Mask)
    {
      /* We already know there's a palette here or this handler
         would never have been installed. */
      palette_t *pal = (palette_t *)WIN_IMAGE(win);
      
      slope_p = pal->slope;
      pal->slope = 
        fexp(flog(old_slope) + (float)(event->xbutton.y-local_y)*0.02);

      if (pal->slope < 1.0/512.0) 
        pal->slope = 1.0/512.0;
      if (pal->slope > 255.0) 
        pal->slope = 255.0;
      
      old_origin -= (pal->slope - slope_p) * 128.0;
      pal->origin = old_origin + event->xbutton.x - local_x;
      
      set_palette (gstate, pal, PAL_MAP);
      
      set_redraw (gstate, win);

      sprintf (line, "palette fiddle %f %f\0", pal->slope, pal->origin);
      GUIaddCommand (line);
      TRANSaddCommand(gstate, line);
    }
  
  return;
}


/* --------------------- GUIpalButton1ReleaseHandler ---------------------- */
static XmxEventHandler (GUIpalButton1ReleaseHandler)
{
  window_t *win = gstate->palette_window;
  palette_t *pal;

  if (win->image == NULL)
    return;

  if (event->xbutton.button == Button1)
    {
      pal = (palette_t *)(win->image);
      pal->probe_value = 0;
      set_redraw (gstate, win);
    }

  return;
}


/* --------------------- GUIpalButton2ReleaseHandler ---------------------- */
static XmxEventHandler (GUIpalButton2ReleaseHandler)
{
  window_t *win = gstate->palette_window;
  
  if (event->xbutton.button == Button2)
    {
      XmxSetUniqid (win->id);
      XmxRemoveEventHandler (pal_dv, Button2MotionMask, GUIpalMotion2Handler, 
                             0);
      XmxRemoveEventHandler (pal_dv, ButtonReleaseMask, 
                             GUIpalButton2ReleaseHandler, 0);

      /* Store the last values for next time around. */
      local_x = event->xbutton.x;
      local_y = pal_dv_height - 1 - event->xbutton.y;
    }

  return;
}


/* ----------------------- GUIpalButtonPressHandler ----------------------- */
static XmxEventHandler (GUIpalButtonPressHandler)
{
  window_t *win = gstate->palette_window;
  palette_t *pal;

  if (WIN_IMAGE(win) == NULL)
    return;

  /* We don't need to do anything special on button1down. */

  if (event->xbutton.button == Button2)
    {
      pal = (palette_t *)WIN_IMAGE(win);
      
      XmxSetUniqid (win->id);
      XmxAddEventHandler (pal_dv, Button2MotionMask, GUIpalMotion2Handler, 0);
      XmxAddEventHandler (pal_dv, ButtonReleaseMask, 
                          GUIpalButton2ReleaseHandler, 0);
      
      local_x = event->xbutton.x;
      local_y = pal_dv_height - 1 - event->xbutton.y;
      old_origin = pal->origin;
      old_slope = pal->slope;
    }

  return;
}



/* ---------------------------- event_palette ----------------------------- */
int event_palette(state_t *state, window_t *win, long dev_id, short data)
{
  return ST_OKAY;
}


/* ---------------------------- notify_palette ---------------------------- */
int notify_palette(state_t * state, window_t * send_win, window_t * recv_win,
                   int message, va_list args)
{
  int status;
    
  status = ST_OKAY;
  
  switch (message) 
    {
    case MSG_NEWDATA:
    case MSG_NEWACT:
      /* Update the palette, if necessary. */
      if ((recv_win != NULL) &&
          (recv_win->image != (void *) send_win->palette)) 
        {
          /* Focus on the palette of the new active Polyview window. */
          recv_win->image = (void *) send_win->palette;
          set_redraw(state, recv_win);
          set_palette(state, (palette_t *) (recv_win->image),
                      PAL_MAP);
          GUIpaletteResetBackdrop (state, send_win);
          GUIpaletteResetPresets (state);
        }
      break;
    case MSG_NEWFRAME:
    case MSG_VIEWCHG:
    case MSG_NEWPICK:
    case MSG_RESET:
      break;
    case MSG_CLOSING:
      /* Disconnect from the active palette and update, if the */
      /* window that is closing is a viewing window. */
      if ((WIN_TYPE(send_win) == POLYVIEW) &&
          (send_win == state->active_windows)) 
        {
          recv_win->image = NULL;
          set_redraw(state, recv_win);
        }
      break;
    default:
      break;
    }
  
  return status;
}



int destroy_palette(state_t *state, window_t *win)
{
  return ST_OKAY;
}



/* ------------------------------------------------------------------------ */
/* ------------------------- GL WIDGET CALLBACKS -------------------------- */
/* ------------------------------------------------------------------------ */

/* This routine converts a widget's background pixel to an RGB color
   suitable for gl. */
unsigned long WidgetBackgroundToGlRgb(Widget widget)
{
  Arg args[10];
  int n;
  Pixel xbg;          /* x background pixel */
  Colormap xcolormap;
  XColor xcolor;
  unsigned long glbg; /* gl bacground color */
  
  /* First get the background pixel from the widget. */
  n = 0;
  XtSetArg(args[n], XtNbackground, &xbg); n++;
  XtGetValues(widget, args, n);
  
  /* Now get the colormap from the top level.  We can't use the widget's
   * colormap because it might not contain the same colors as the
   * colors that the background color were allocated from, so we use it's
   * parent. */
  n = 0;
  XtSetArg(args[n], XtNcolormap, &xcolormap); n++;
  XtGetValues(XtParent(widget), args, n);
  
  /* Now obtain RGB values */
  xcolor.flags = DoRed | DoGreen | DoBlue;
  xcolor.pixel = xbg;
  XQueryColor (XtDisplay(widget), xcolormap, &xcolor);
  
  /* Now pack into an RGB color suitable for GL.  The format is 0x00BBGGRR
   * Since the x format has colors values from 0 to 65535 first shift the
   * x values right by 8 and then shift left to fit into the rgb color.*/
  glbg = (xcolor.red >> 8) + ((xcolor.green >> 8) << 8) +
    ((xcolor.blue >> 8) << 16);
  return (glbg);
}


/* --------------------------- GUIpalRedrawCB --------------------------- */
XmxCallback (GUIpalRedrawCB)
{
  set_redraw (gstate, gstate->palette_window);

  return;
}


/* --------------------------- GUIpalGinitCB ---------------------------- */
XmxCallback (GUIpalGinitCB)
{
  window_t *win = gstate->palette_window;

  /* ---> This callback is supposed to be hit when the Widget
          is first realized. */
  
  /* Focus on this window. */
  XmxWinset (w);

  /* Set up initial parameters. */
  doublebuffer ();
  RGBmode ();

  /* Set up zbuffering. */
  zbuffer(TRUE);

  XmxInstallColormaps (gstate->base, pal_dv);
  
  if (!(gstate->gl_initialized))
    {
      gstate->gl_initialized = 1;
      
      /* Right now lighting and patterns init is here since we know 
         this is the first GL dialog to be opened. */
      init_lighting ();
      init_patterns ();
      init_zbuffer ();
    }

  /* Go grab the background color. */
  pal_bg = WidgetBackgroundToGlRgb (pal_dv);
  
  /* This should be all that's needed; the rest of the work takes
     place in redraw_polyview, right? */

  /* We get a redraw callback here. */

  return;
}




/* --------------------------- GUIpalResizeCB --------------------------- */
XmxCallback (GUIpalResizeCB)
{
  window_t *win = gstate->palette_window;

  /* Do a GL focus on the window as we get it from the callback. */
  XmxWinset (w);

  /* Store the new width and height in the window image. */
#ifdef __sgi
  pal_dv_width = ((GlxDrawCallbackStruct *)call_data)->width;
  pal_dv_height = ((GlxDrawCallbackStruct *)call_data)->height;
#endif
  
#ifdef _IBMR2
  {
    Dimension x, y;
    XmxSetArg (XmNwidth, &x);
    XmxSetArg (XmNheight, &y);
    XtGetValues (w, Xmx_wargs, Xmx_n);
    pal_dv_width = x;
    pal_dv_height = y;
  }
#endif

  /* Go update the rest of Calvert's data structures. */
  win->width = pal_dv_width;
  win->height = pal_dv_height;
  win->cx = win->width / 2;
  win->cy = win->height / 2;
  
  viewport
    (0, (Screencoord)pal_dv_width - 1, 0, (Screencoord)pal_dv_height - 1);
  ortho2 (0.0, (float)(pal_dv_width - 1), 0.0, (float)(pal_dv_height - 1));
  
  /* We get a redraw callback here. */
  
  return;
}


XmxCallback (GUIpaletteFSBCallback)
{
  char *fname;
  char *what;
  char line[MAXLINELEN];

  PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/tennis.aiff");
  
  fname = (char *)PVMALLOC (128 * sizeof (char));

  XmStringGetLtoR (((XmFileSelectionBoxCallbackStruct *)call_data)->value,
                   XmSTRING_DEFAULT_CHARSET,
                   &fname);

  switch (XmxExtractToken ((int)client_data))
    {
    case PV_PALETTE_LOAD_HDF:
      what = "load hdf";
      break;
    case PV_PALETTE_LOAD_RAW:
      what = "load raw";
      break;
    case PV_PALETTE_SAVE_HDF:
      what = "save hdf";
      break;
    case PV_PALETTE_SAVE_RAW:
      what = "save raw";
      break;
    }

  sprintf (line, "palette %s %s\0", what, fname);
  parse_line (gstate, line, TRUE);
  
  free (fname);
  XtUnmanageChild (w);

  return;
}



/* ------------------------- GUIpaletteMenubarCB -------------------------- */
XmxCallback (GUIpaletteMenubarCB)
{
  switch (XmxExtractToken ((int)client_data))
    {
    case PV_PALETTE_LOAD_HDF:
      if (load_hdf == NULL)
        load_hdf = XmxMakeFileSBDialog
          (gstate->palette_window->base, "Load HDF Palette",
           "Enter filename of HDF palette:",
           GUIpaletteFSBCallback, PV_PALETTE_LOAD_HDF);
      else
        XmFileSelectionDoSearch (load_hdf, NULL);
      XmxManageRemanage (load_hdf);
      break;
    case PV_PALETTE_LOAD_RAW:
      if (load_raw == NULL)
        load_raw = XmxMakeFileSBDialog
          (gstate->palette_window->base, "Load HDF Palette",
           "Enter filename of HDF palette:",
           GUIpaletteFSBCallback, PV_PALETTE_LOAD_RAW);
      else
        XmFileSelectionDoSearch (load_raw, NULL);
      XmxManageRemanage (load_raw);
      break;
    case PV_PALETTE_SAVE_HDF:
      if (save_hdf == NULL)
        save_hdf = XmxMakeFileSBDialog
          (gstate->palette_window->base, "Load HDF Palette",
           "Enter filename of HDF palette:",
           GUIpaletteFSBCallback, PV_PALETTE_SAVE_HDF);
      else
        XmFileSelectionDoSearch (save_hdf, NULL);
      XmxManageRemanage (save_hdf);
      break;
    case PV_PALETTE_SAVE_RAW:
      if (save_raw == NULL)
        save_raw = XmxMakeFileSBDialog
          (gstate->palette_window->base, "Load HDF Palette",
           "Enter filename of HDF palette:",
           GUIpaletteFSBCallback, PV_PALETTE_SAVE_RAW);
      else
        XmFileSelectionDoSearch (save_raw, NULL);
      XmxManageRemanage (save_raw);
      break;
    case PV_PALETTE_RESET:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/tennis.aiff");
      parse_line (gstate, "palette reset", TRUE);
      break;
    case PV_PALETTE_CLOSE_WINDOW:
      parse_line (gstate, "window close palette", TRUE);
      break;

    case PV_PALETTE_PRESETS_RAINBOW:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use rainbow", TRUE);
      break;
    case PV_PALETTE_PRESETS_WTOB:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use w_to_b", TRUE);
      break;
    case PV_PALETTE_PRESETS_BTOW:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use b_to_w", TRUE);
      break;
    case PV_PALETTE_PRESETS_RGB:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use rgb", TRUE);
      break;
    case PV_PALETTE_PRESETS_YELLOW:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use allyellow", TRUE);
      break;
    case PV_PALETTE_PRESETS_REDBLUE:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use redblue", TRUE);
      break;
    case PV_PALETTE_PRESETS_WHITEBLUE:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use whiteblue", TRUE);
      break;
    case PV_PALETTE_PRESETS_YELLOWRED:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "palette use yellowred", TRUE);
      break;

    case PV_PALETTE_BACKG_BLACK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop solid black", TRUE);
      break;
    case PV_PALETTE_BACKG_WHITE:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop solid white", TRUE);
      break;
    case PV_PALETTE_BACKG_GREY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop solid grey", TRUE);
      break;
    case PV_PALETTE_BACKG_BLUE:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop solid blue", TRUE);
      break;
    case PV_PALETTE_BACKG_RED:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop solid red", TRUE);
      break;

    case PV_PALETTE_BACKG_RGBY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop shaded rgby", TRUE);
      break;
    case PV_PALETTE_BACKG_RGBLUE:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop shaded rgblue", TRUE);
      break;
    case PV_PALETTE_BACKG_RBBLACK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop shaded rbblack", TRUE);
      break;
      
    case PV_PALETTE_BACKG_BLUEHORIZ:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop horizon blue", TRUE);
      break;
    case PV_PALETTE_BACKG_REDHORIZ:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop horizon red", TRUE);
      break;
    case PV_PALETTE_BACKG_GREENHORIZ:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop horizon green", TRUE);
      break;
    case PV_PALETTE_BACKG_YELLOWHORIZ:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop horizon yellow", TRUE);
      break;
      
    case PV_PALETTE_BACKG_BLUESKY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop sky blue", TRUE);
      break;
    case PV_PALETTE_BACKG_REDSKY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop sky red", TRUE);
      break;
    case PV_PALETTE_BACKG_GREENSKY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop sky green", TRUE);
      break;
    case PV_PALETTE_BACKG_YELLOWSKY:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop sky yellow", TRUE);
      break;
      
    case PV_PALETTE_BACKG_BLUESTREAK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop streak blue", TRUE);
      break;
    case PV_PALETTE_BACKG_REDSTREAK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop streak red", TRUE);
      break;
    case PV_PALETTE_BACKG_GREENSTREAK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop streak green", TRUE);
      break;
    case PV_PALETTE_BACKG_YELLOWSTREAK:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop streak yellow", TRUE);
      break;

    case PV_PALETTE_BACKG_BLUESMEAR:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop smear blue", TRUE);
      break;
    case PV_PALETTE_BACKG_REDSMEAR:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop smear red", TRUE);
      break;
    case PV_PALETTE_BACKG_GREENSMEAR:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop smear green", TRUE);
      break;
    case PV_PALETTE_BACKG_YELLOWSMEAR:
      PLAYAFILE ("/usr/lib/sounds/prosonus/sfx/door_slam.aiff");
      parse_line (gstate, "backdrop smear yellow", TRUE);
      break;

    case PV_PALETTE_HELP_ONWINDOW:
      GUIdoHelp (PV_PALETTE_HELP_ONWINDOW);
      break;
    default:
      break;
    }

  return;
}



/* ----------------------------- MENUBAR DATA ----------------------------- */
static XmxMenubarStruct PaletteFileMenu[] =
{
  { "Load HDF...",     'L', GUIpaletteMenubarCB, PV_PALETTE_LOAD_HDF },
  { "Load Raw...",     'o', GUIpaletteMenubarCB, PV_PALETTE_LOAD_RAW },
  { "----" },
  { "Save HDF...",     'S', GUIpaletteMenubarCB, PV_PALETTE_SAVE_HDF },
  { "Save Raw...",     'a', GUIpaletteMenubarCB, PV_PALETTE_SAVE_RAW },
  { "----" },
  { "Reset",           'R', GUIpaletteMenubarCB, PV_PALETTE_RESET },
  { "----" },
  { "Close Window",    'C', GUIpaletteMenubarCB, PV_PALETTE_CLOSE_WINDOW },
  { "Exit Program...", 'E', GUImainMenubarCB,    PV_EXIT_PROGRAM },
  { NULL },
};

static XmxMenubarStruct PaletteWindowsMenu[] =
{
  { "Open Polyview...", 'O', GUImainMenubarCB, PV_OPEN_POLYVIEW },
  { "Open Palette...",  'P', GUImainMenubarCB, PV_OPEN_PALETTE },
  { "Open Info...",     'I', GUImainMenubarCB, PV_OPEN_INFO },
  { "Open Time...",     'T', GUImainMenubarCB, PV_OPEN_TIME },
  { NULL },
};

static XmxMenubarStruct PalettePresetsMenu[] =
{
  { "<Rainbow",        'R', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_RAINBOW },
  { "<Red to Blue",    'd', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_REDBLUE },
  { "<White to Blue",  'h', GUIpaletteMenubarCB, 
      PV_PALETTE_PRESETS_WHITEBLUE },
  { "<Yellow to Red",  'e', GUIpaletteMenubarCB, 
      PV_PALETTE_PRESETS_YELLOWRED },
  { "<All Yellow",     'Y', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_YELLOW },
  { "<White to Black", 'W', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_WTOB },
  { "<Black to White", 'B', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_BTOW },
  { "<RGB",            'G', GUIpaletteMenubarCB, PV_PALETTE_PRESETS_RGB },
  { NULL },
};

static XmxMenubarStruct solidMenu[] =
{
  { "<Black", 'B',  GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLACK },
  { "<White", 'W',  GUIpaletteMenubarCB, PV_PALETTE_BACKG_WHITE },
  { "<Grey",  'G',  GUIpaletteMenubarCB, PV_PALETTE_BACKG_GREY },
  { "<Blue",  'l',  GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLUE },
  { "<Red",   'R',  GUIpaletteMenubarCB, PV_PALETTE_BACKG_RED },
  { NULL },
};

static XmxMenubarStruct shadedMenu[] =
{
  { "<RGBY",    NULL, GUIpaletteMenubarCB, PV_PALETTE_BACKG_RGBY },
  { "<RGB",     NULL, GUIpaletteMenubarCB, PV_PALETTE_BACKG_RGBLUE },
  { "<RBBlack", NULL, GUIpaletteMenubarCB, PV_PALETTE_BACKG_RBBLACK },
  { NULL },
};

static XmxMenubarStruct horizonMenu[] =
{
  { "<BlueHorizon",   'B', GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLUEHORIZ },
  { "<RedHorizon",    'R', GUIpaletteMenubarCB, PV_PALETTE_BACKG_REDHORIZ },
  { "<GreenHorizon",  'G', GUIpaletteMenubarCB, PV_PALETTE_BACKG_GREENHORIZ },
  { "<YellowHorizon", 'Y', GUIpaletteMenubarCB, PV_PALETTE_BACKG_YELLOWHORIZ },
  { NULL },
};

static XmxMenubarStruct skyMenu[] =
{
  { "<BlueSky",   'B', GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLUESKY },
  { "<RedSky",    'R', GUIpaletteMenubarCB, PV_PALETTE_BACKG_REDSKY },
  { "<GreenSky",  'G', GUIpaletteMenubarCB, PV_PALETTE_BACKG_GREENSKY },
  { "<YellowSky", 'Y', GUIpaletteMenubarCB, PV_PALETTE_BACKG_YELLOWSKY },
  { NULL },
};

static XmxMenubarStruct streakMenu[] =
{
  { "<BlueStreak",   'B', GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLUESTREAK },
  { "<RedStreak",    'R', GUIpaletteMenubarCB, PV_PALETTE_BACKG_REDSTREAK },
  { "<GreenStreak",  'G', GUIpaletteMenubarCB, PV_PALETTE_BACKG_GREENSTREAK },
  { "<YellowStreak", 'Y', GUIpaletteMenubarCB, PV_PALETTE_BACKG_YELLOWSTREAK },
  { NULL },
};

static XmxMenubarStruct smearMenu[] =
{
  { "<BlueSmear",   'B', GUIpaletteMenubarCB, PV_PALETTE_BACKG_BLUESMEAR },
  { "<RedSmear",    'R', GUIpaletteMenubarCB, PV_PALETTE_BACKG_REDSMEAR },
  { "<GreenSmear",  'G', GUIpaletteMenubarCB, PV_PALETTE_BACKG_GREENSMEAR },
  { "<YellowSmear", 'Y', GUIpaletteMenubarCB, PV_PALETTE_BACKG_YELLOWSMEAR },
  { NULL },
};

static XmxMenubarStruct PaletteBackgroundMenu[] =
{
  { "Solid",   'S', NULL, NULL, solidMenu },
  { "Shaded",  'd', NULL, NULL, shadedMenu },
  { "Horizon", 'H', NULL, NULL, horizonMenu },
  { "Sky",     'k', NULL, NULL, skyMenu },
  { "Streak",  't', NULL, NULL, streakMenu },
  { "Smear",   'm', NULL, NULL, smearMenu },
  { NULL },
};

static XmxMenubarStruct PaletteHelpMenu[] =
{
  { "About",       'A', GUImainMenubarCB,    PV_HELP_ABOUT },
  { "On Commands", 'C', GUImainMenubarCB, PV_HELP_ONCOMMANDS},
  { "On Window",   'W', GUIpaletteMenubarCB, PV_PALETTE_HELP_ONWINDOW },
  { NULL },
};

static XmxMenubarStruct PaletteMenu[] = 
{
  { "File",       'F', NULL, NULL, PaletteFileMenu },
  { "Windows",    'W', NULL, NULL, PaletteWindowsMenu },
  { "Palette",    'P', NULL, NULL, PalettePresetsMenu },
  { "Backdrop",   'B', NULL, NULL, PaletteBackgroundMenu },
  { "Help",       'H', NULL, NULL, PaletteHelpMenu },
  { NULL },
};


/* ------------------------------------------------------------------------ */

#ifdef __sgi
static GLXconfig palGlxConfig[] =
{
  { GLX_NORMAL, GLX_DOUBLE, TRUE },
  { GLX_NORMAL, GLX_RGB, TRUE },
  { 0, 0, 0 },
};
#endif

/* This is the entry point from pvfuncs.c. */
int post_palette(state_t *state)
{
  window_t *win;
  Widget drawing_frame;

  if (STA_PALETTE(state) == NULL) 
    {
      /* Create the palette win. */
      /* win_geometry_auto(state, PREF); */
      STA_PALETTE(state) =
        win = add_window(state, "palette", PALETTE,
                         init_palette, redraw_palette,
                         animate_palette, event_palette,
                         notify_palette, destroy_palette);
      
      /* If the window was not created, abort the program with an */
      /* error. */
      assert (win != NULL);

      win->id = 0;
      XmxSetUniqid (win->id);
      win->base = XmxMakeFormDialog 
        (gstate->base, "Polyview 3.1: Palette Window");
      
      /* Make the menubar first. */
      pal_menubar = XmxRMakeMenubar (win->base, PaletteMenu);
      
      /* Now make the drawing area (later, GL widget). */
      drawing_frame = XmxMakeFrame (win->base, XmxShadowOut);
      pal_dv_width = win->width = 390;
      pal_dv_height = win->height = 160;
      win->cx = win->width / 2;
      win->cy = win->height / 2;
#ifdef __sgi
      pal_dv = XmxMakeDrawingVolume 
        (drawing_frame, pal_dv_width, pal_dv_height, palGlxConfig,
         GUIpalRedrawCB, GUIpalResizeCB, GUIpalGinitCB);
#endif

#ifdef _IBMR2
      pal_dv = XmxMakeDrawingVolume 
        (drawing_frame, pal_dv_width, pal_dv_height,
         GUIpalRedrawCB, GUIpalResizeCB, GUIpalGinitCB);
#endif

      /* Children of base. */
      XmxSetConstraints 
        (pal_menubar->base, XmATTACH_FORM, XmATTACH_NONE, XmATTACH_FORM, 
         XmATTACH_FORM, NULL, NULL, NULL, NULL);
      XmxSetConstraints 
        (drawing_frame, XmATTACH_WIDGET, XmATTACH_FORM, XmATTACH_FORM, 
         XmATTACH_FORM, pal_menubar->base, NULL, NULL, NULL);
      
      /* File selection boxes aren't created initially. */
      load_hdf = load_raw = save_hdf = save_raw = NULL;
      
      /* Attach the active window's palette to this window. */
      if (state->active_windows == NULL)
        WIN_IMAGE(win) = NULL;
      else
        WIN_IMAGE(win) = (void *)WIN_PALETTE(state->active_windows);

      GUIpaletteResetPresets (gstate);
      GUIpaletteResetBackdrop (gstate, state->active_windows);

      /* Attach event handlers. */
      XmxAddEventHandler (pal_dv, Button1MotionMask, GUIpalMotion1Handler, 0);
      XmxAddEventHandler (pal_dv, ButtonReleaseMask, 
                          GUIpalButton1ReleaseHandler, 0);
      XmxAddEventHandler (pal_dv, ButtonPressMask, 
                          GUIpalButtonPressHandler, 0);
    }

  XmxManageRemanage (STA_PALETTE(state)->base);
  return ST_OKAY;
}



/* ----------------------------- set_palette ------------------------------ */
int set_palette(state_t *state, palette_t *palette, int type)
{
  short (*master)[3];
  short (*display)[3];
  /* Color's red, green, and blue components (0..255) */
  float r, g, b;
  float slope, origin;
  int base, i;
  
  /* Color's white component (0..255) */
  float w;
  /* Mapped value for color "fiddling" */
  int x;
  
  window_t *win;
  
  /* Initialize the master and display palette pointers. */
  master = palette->master_palette;
  display = palette->display_palette;
    
  /* Set up the master palette with the specified set of colors or */
  /* remap the colors to the display palette. */
  switch (type) 
    {
    case PAL_RGB:
      /* This is a _really_ stupid colormap. */
      for (i = 0; i < 256; i++)
        {
          /* The color components map to the value of the */
          /* appropriate bit fields: */
          /*	76543210 */
          /*	BBGGGRRR */
          /* Yeah, sure, but what the hell for? */
          master[i][R] = (short) (i & 0x07) << 5;
          master[i][G] = (short) (i & 0x38) << 2;
          master[i][B] = (short) i & 0xc0;
        }
      break;
    case PAL_W_TO_B:
      /* Initialize the white component for the "black" end of */
      /* the spectrum. */
      w = 0.0;
      for (i = 255; i >= 0; i--) 
        {
          master[i][R] = (short) w;
          master[i][G] = (short) w;
          master[i][B] = (short) w;
          w += 1.0;
        }
      break;
    case PAL_B_TO_W:
      /* Initialize the white component for the "black" end of */
      /* the spectrum. */
      w = 0.0;
      for (i = 0; i < 256; i++) 
        {
          master[i][R] = (short) w;
          master[i][G] = (short) w;
          master[i][B] = (short) w;
          w += 1.0;
        }
      break;
    case PAL_ALLYELLOW:
      /* This is for lighting. */
      for (i = 0; i < 256; i++)
        {
          master[i][R] = master[i][G] = 0xcf;
          master[i][B] = 0x3f;
        }
      break;
    case PAL_REDBLUE:
      for (i = 0; i < 256; i++) 
        {
          master[i][R] = i;
          master[i][G] = 0;
          master[i][B] = 255 - (short) i;
        }
      break;
    case PAL_WHITEBLUE:
      for (i = 0; i < 256; i++) 
        {
          master[i][R] = MAX(i-30, 0);
          master[i][G] = MAX(i-30, 0);
          master[i][B] = 255;
        }
      break;
    case PAL_YELLOWRED:
      for (i = 0; i < 256; i++) 
        {
          master[i][R] = 255;
          master[i][G] = MAX(i-30, 0);
          master[i][B] = 0;
        }
      break;
    case PAL_RAINBOW:
      /* Use the default rainbow palette */
      /* Initialize the color components for the rainbow palette */
      r = 129.0;
      g = 0.0;
      b = 255.0;
      
      /* Violet to blue.  Green, blue constant, red changes */
      for (i = 0; i < 29; i++) 
        {
          master[i][R] = (short) r;
          master[i][G] = (short) g;
          master[i][B] = (short) b;
          r -= (130.0/29.0);
        }
      
      r = 0.0;
      g = 2.036;
      /* Blue to light blue.  Red, blue constant, green changes */
      for (; i < 86; i++) 
        {
          master[i][R] = (short) r;
          master[i][G] = (short) g;
          master[i][B] = (short) b;
          g += (257.0/57.0);
        }
      
      g = 255.0;
      b = 251.5;
      /* Light blue to green.  Red, green constant, blue changes */
      for (; i < 142; i++) 
        {
          master[i][R] = (short) r;
          master[i][G] = (short) g;
          master[i][B] = (short) b;
          b -= (247.0/55.0);
        }
      
      b = 0.0;
      r = 1.27;
      /* Green to yellow.  Green, blue constant, red changes */
      for (; i < 199; i++) 
        {
          master[i][R] = (short) r;
          master[i][G] = (short) g;
          master[i][B] = (short) b;
          r += (252.5/56.0);
        }
      
      r = 255.0;
      g = 252.5;
      /* Yellow to red.  Red, blue constant, green changes */
      for (; i < 256; i++) 
        {
          master[i][R] = (short) r;
          master[i][G] = (short) g;
          master[i][B] = (short) b;
          g -= (252.0/56.0);
        }
      
      break;
    case PAL_FILE:
    case PAL_RESET:
      palette->origin = 0.0;
      palette->slope = 1.0;
    case PAL_MAP:
      origin = palette->origin;
      slope = palette->slope;
      for (i = 0; i < 256; i++) 
        {
          /* Calculate the mapping function from the master */
          /* palette to the display palette */
          x =  (((float)i - origin) / slope);
          if (x < 0) x = 0;
          if (x > 255) x = 255;
          display[i][R] = master[x][R];
          display[i][G] = master[x][G];
          display[i][B] = master[x][B];
        }
      break;
    default:
      stprintf(state, "Unknown color map type.");
      break;
    }
  
  /* Map the colors from the master palette to the display palette */
  /* if any change has taken place. */
  if (type != PAL_MAP)
    for (i = 0; i < 256; i++) 
      {
        display[i][R] = master[i][R];
        display[i][G] = master[i][G];
        display[i][B] = master[i][B];
      }
  
  /* Copy the palette into the system palette for display */
  /* And into the master palette array for reference */
  base = palette->base;
  
  for (i = 0; i < 256; i++, display++) 
    {
      state->pal[i][R] = display[0][R];
      state->pal[i][G] = display[0][G];
      state->pal[i][B] = display[0][B];
    }
  
  /* Go get a redraw for windows (kinda brutal, but probably
     necessary). */
  FOR_ALL_WINDOWS (state, win)
    set_redraw(state, win);

  /* process_events(state); */

  /* This just plain don't work right. */
  if (type != last_set_palette && type != PAL_MAP &&
      type != PAL_RESET)
    {
      /* We need to reset the presets menubar. */
      last_set_palette = type;
      GUIpaletteResetPresets (state);
    }
  
  return ST_OKAY;
}
